"""
Abstract base class for tracker jobs
"""

import abc
import builtins
import collections
import functools
import pathlib

from ... import errors, jobs, utils

import logging  # isort:skip
_log = logging.getLogger(__name__)


class TrackerJobsBase(abc.ABC):
    """
    Base class for tracker-specific :class:`jobs <upsies.jobs.base.JobBase>`

    This base class defines general-purpose jobs that can be used by subclasses
    by returning them in their :attr:`jobs_before_upload` or
    :attr:`jobs_after_upload` attributes. It also provides all objects that are
    needed by any one of those jobs.

    Job instances are provided as :func:`functools.cached_property`, i.e. jobs
    are created only once per session.

    Subclasses that need to run background tasks should pass them to
    :meth:`.JobBase.attach_task` or to :meth:`.TrackerBase.attach_task`.

    For a description of the arguments see the corresponding properties.
    """

    def __init__(self, *, content_path, tracker,
                 reuse_torrent_path=None, exclude_files=(),
                 btclient=None, torrent_destination=None,
                 screenshots_optimization=None, image_hosts=None,
                 options=None, common_job_args=None):
        self._content_path = content_path
        self._reuse_torrent_path = reuse_torrent_path
        self._tracker = tracker
        self._screenshots_optimization = screenshots_optimization
        self._image_hosts = image_hosts
        self._btclient = btclient
        self._torrent_destination = torrent_destination
        self._exclude_files = exclude_files
        self._common_job_args = common_job_args or {}
        self._options = options or {}
        self._signal = utils.signal.Signal('warning', 'error', 'exception')

    @property
    def content_path(self):
        """
        Content path to generate metadata for

        This is the same object that was passed as initialization argument.
        """
        return self._content_path

    @property
    def tracker(self):
        """
        :class:`~.trackers.base.TrackerBase` subclass

        This is the same object that was passed as initialization argument.
        """
        return self._tracker

    @property
    def reuse_torrent_path(self):
        """
        Path to an existing torrent file that matches :attr:`content_path`

        See :func:`.torrent.create`.
        """
        return self._reuse_torrent_path

    @property
    def torrent_destination(self):
        """
        Where to copy the generated torrent file to or `None`

        This is the same object that was passed as initialization argument.
        """
        return self._torrent_destination

    @property
    def exclude_files(self):
        """
        Sequence of glob and regular expression patterns to exclude from the
        generated torrent

        See the ``exclude_files`` argument of
        :meth:`.CreateTorrentJob.initialize`.

        This should also be used by :attr:`screenshots_job` to avoid making
        screenshots from files that aren't in the torrent.
        """
        return self._exclude_files

    @property
    def options(self):
        """
        Configuration options provided by the user

        This is the same object that was passed as initialization argument.
        """
        return self._options

    @property
    def image_hosts(self):
        """
        Sequence of :class:`~.base.ImageHostBase` instances or `None`

        This is the same object that was passed as initialization argument.
        """
        return self._image_hosts

    @property
    def btclient(self):
        """
        :class:`~.btclient.BtClient` instance or `None`

        This is the same object that was passed as initialization argument.
        """
        return self._btclient

    def common_job_args(self, **overload):
        """
        Keyword arguments that are passed to all jobs or empty `dict`

        :param overload: Keyword arguments add or replace values from the
            initialization argument
        """
        return {
            **self._common_job_args,
            **overload,
        }

    @property
    @abc.abstractmethod
    def jobs_before_upload(self):
        """
        Sequence of jobs that need to finish before :meth:`~.TrackerBase.upload` can
        be called
        """

    @functools.cached_property
    def jobs_after_upload(self):
        """
        Sequence of jobs that are started after :meth:`~.TrackerBase.upload`
        finished

        .. note:: Jobs returned by this class should have
                  :attr:`~.JobBase.autostart` set to `False` or they will be
                  started before submission is attempted.

        By default, this returns :attr:`add_torrent_job` and
        :attr:`copy_torrent_job`.
        """
        return (
            self.add_torrent_job,
            self.copy_torrent_job,
        )

    @property
    def isolated_jobs(self):
        """
        Sequence of job names (e.g. ``"imdb_job"``) that were singled out by the
        user (e.g. with a CLI argument) to create only a subset of the usual
        metadata

        If this sequence is empty, all jobs in :attr:`jobs_before_upload` and
        :attr:`jobs_after_upload` are enabled.
        """
        return ()

    def get_job_and_dependencies(self, *jobs):
        """
        Combine all `jobs` and their dependencies recursively into flat list

        :param jobs: :class:`~.JobBase` instances

        Dependencies are gathered from each job's :attr:`~.JobBase.prejobs` and
        :meth:`~.JobBase.presignal`\\ s.

        .. warning:: This is not a foolproof way to find all dependencies.

                     :class:`~.QueueJobBase` instances, for example, don't where
                     their input comes from, so they don't know anything about
                     which jobs they depend on. There are also
                     :attr:`~.JobBase.precondition`\\ s, which can define
                     arbitrary requirements.

        :return: Sequence of :class:`~.JobBase` instances
        """
        all_jobs = []

        def add_job(job):
            if job not in all_jobs:
                all_jobs.append(job)

        def add_prejobs(job):
            for prejob in job.prejobs:
                add_job(prejob)
                add_prejobs(prejob)
                add_jobs_from_presignals(prejob)

        def add_jobs_from_presignals(job):
            for presignal_job in job.presignals:
                add_job(presignal_job)
                add_prejobs(presignal_job)
                add_jobs_from_presignals(presignal_job)

        for j in jobs:
            add_job(j)
            add_prejobs(j)
            add_jobs_from_presignals(j)

        return all_jobs

    @property
    def submission_ok(self):
        """
        Whether the created metadata should be submitted

        The base class implementation returns `False` if there are any
        :attr:`isolated_jobs`. Otherwise, it returns `True` only if all
        :attr:`jobs_before_upload` have an :attr:`~.base.JobBase.exit_code` of
        ``0`` or a falsy :attr:`~.base.JobBase.is_enabled` value.

        Subclasses should always call the parent class implementation to ensure
        all metadata was created successfully.
        """
        if self.isolated_jobs:
            # If some jobs are disabled, required metadata is missing and we
            # can't submit
            return False
        else:
            enabled_jobs_before_upload = tuple(
                job for job in self.jobs_before_upload
                if job and job.is_enabled
            )
            enabled_jobs_succeeded = all(
                job.exit_code == 0
                for job in enabled_jobs_before_upload
            )
            return bool(
                enabled_jobs_before_upload
                and enabled_jobs_succeeded
            )

    @property
    def signal(self):
        """
        :class:`~.signal.Signal` instance with the signals ``warning``, ``error``
        and ``exception``
        """
        return self._signal

    def warn(self, warning):
        """
        Emit ``warning`` signal (see :attr:`signal`)

        Emit a warning for any non-critical issue that the user can choose to
        ignore or fix.
        """
        self.signal.emit('warning', warning)

    def error(self, error):
        """
        Emit ``error`` signal (see :attr:`signal`)

        Emit an error for any critical but expected issue that can't be
        recovered from (e.g. I/O error).
        """
        self.signal.emit('error', error)

    def exception(self, exception):
        """
        Emit ``exception`` signal (see :attr:`signal`)

        Emit an exception for any critical and unexpected issue that should be
        reported as a bug.
        """
        self.signal.emit('exception', exception)

    @functools.cached_property
    def imdb(self):
        """:class:`~.webdbs.imdb.ImdbApi` instance"""
        return utils.webdbs.webdb('imdb')

    @functools.cached_property
    def tmdb(self):
        """:class:`~.webdbs.tmdb.TmdbApi` instance"""
        return utils.webdbs.webdb('tmdb')

    @functools.cached_property
    def tvmaze(self):
        """:class:`~.webdbs.tvmaze.TvmazeApi` instance"""
        return utils.webdbs.webdb('tvmaze')

    def get_job_name(self, name):
        """
        Return job name that is unique for this tracker

        It's important for tracker jobs to have unique names to avoid re-using
        cached output from another tracker's job with the same name.

        Standard jobs have names so that cached output will be re-used by other
        trackers if possible. This function is mainly for unique and custom jobs
        that are only used for one tracker but might share the same name with
        other trackers.
        """
        suffix = f'.{self.tracker.name}'
        if name.endswith(suffix):
            return name
        else:
            return f'{name}{suffix}'

    @functools.cached_property
    def create_torrent_job(self):
        """:class:`~.jobs.torrent.CreateTorrentJob` instance"""
        return jobs.torrent.CreateTorrentJob(
            content_path=self.content_path,
            reuse_torrent_path=self.reuse_torrent_path,
            tracker=self.tracker,
            exclude_files=self.exclude_files,
            precondition=self.make_precondition('create_torrent_job'),
            **self.common_job_args(),
        )

    @functools.cached_property
    def add_torrent_job(self):
        """:class:`~.jobs.torrent.AddTorrentJob` instance"""
        if self._btclient and self.create_torrent_job:
            add_torrent_job = jobs.torrent.AddTorrentJob(
                autostart=False,
                btclient=self._btclient,
                precondition=self.make_precondition('add_torrent_job'),
                **self.common_job_args(),
            )

            # Pass CreateTorrentJob output to AddTorrentJob input.
            self.create_torrent_job.signal.register('output', add_torrent_job.enqueue)

            # Tell AddTorrentJob to finish the current upload and then finish.
            self.create_torrent_job.signal.register('finished', self.finalize_add_torrent_job)

            return add_torrent_job

    def finalize_add_torrent_job(self, _):
        self.add_torrent_job.finalize()

    @functools.cached_property
    def copy_torrent_job(self):
        """:class:`~.jobs.torrent.CopyTorrentJob` instance"""
        if self.torrent_destination and self.create_torrent_job:
            copy_torrent_job = jobs.torrent.CopyTorrentJob(
                autostart=False,
                destination=self.torrent_destination,
                precondition=self.make_precondition('copy_torrent_job'),
                **self.common_job_args(),
            )
            # Pass CreateTorrentJob output to CopyTorrentJob input.
            self.create_torrent_job.signal.register('output', copy_torrent_job.enqueue)
            # Tell CopyTorrentJob to finish when CreateTorrentJob is done.
            self.create_torrent_job.signal.register('finished', self.finalize_copy_torrent_job)
            return copy_torrent_job

    def finalize_copy_torrent_job(self, _):
        self.copy_torrent_job.finalize()

    @property
    def torrent_filepath(self):
        """Local path to the torrent file created by :attr:`create_torrent_job`"""
        return self.get_job_output(self.create_torrent_job, slice=0)

    @functools.cached_property
    def subtitles(self):
        """Sequence of :class:`~.subtitle.Subtitle` objects for :attr:`content_path`"""
        return utils.subtitle.get_subtitles(self.content_path)

    @functools.cached_property
    def release_name(self):
        """
        :class:`~.release.ReleaseName` instance with
        :attr:`release_name_translation` applied
        """
        return utils.release.ReleaseName(
            path=self.content_path,
            translate=self.release_name_translation,
            separator=self.release_name_separator,
        )

    release_name_separator = None
    """See :attr:`.ReleaseName.separator`"""

    release_name_translation = {}
    """See ``translate`` argument of :attr:`~.utils.release.ReleaseName`"""

    @functools.cached_property
    def release_name_job(self):
        """
        :class:`~.jobs.dialog.TextFieldJob` instance with text set to
        :attr:`release_name`

        The text is automatically updated when :attr:`imdb_job` sends an ID.
        """
        return jobs.dialog.TextFieldJob(
            name=self.get_job_name('release-name'),
            label='Release Name',
            text=str(self.release_name),
            callbacks={
                'output': self.release_name.set_release_info,
            },
            precondition=self.make_precondition('release_name_job'),
            **self.common_job_args(),
        )

    async def update_release_name_from(self, webdb, webdb_id):
        """
        Update :attr:`release_name_job` with web DB information

        :param webdb: :class:`~.webdbs.base.WebDbApiBase` instance
        :param webdb_id: ID for `webdb`

        This is a convenience wrapper around :meth:`.ReleaseName.fetch_info` and
        :meth:`.TextFieldJob.fetch_text`.
        """
        await self.release_name_job.fetch_text(
            coro=self.release_name.fetch_info(webdb=webdb, webdb_id=webdb_id),
            nonfatal_exceptions=(errors.RequestError,),
            default=str(self.release_name),
        )
        _log.debug('Updated release name: %s', self.release_name)

    @functools.cached_property
    def imdb_job(self):
        """:class:`~.jobs.webdb.WebDbSearchJob` instance"""
        return jobs.webdb.WebDbSearchJob(
            query=self.content_path,
            db=self.imdb,
            callbacks={
                'output': self._handle_imdb_id,
            },
            precondition=self.make_precondition('imdb_job'),
            **self.common_job_args(),
        )

    def _handle_imdb_id(self, imdb_id):
        # Update other webdb queries with IMDb info
        self.tracker.attach_task(self._propagate_webdb_info(self.imdb, imdb_id))

        # Upload poster image from appropriate webdb
        if self._poster_webdb is self.imdb:
            self._start_poster_job(self.imdb, imdb_id)

    @property
    def imdb_id(self):
        """IMDb ID if :attr:`imdb_job` is finished or `None`"""
        return self._get_webdb_id_from_job(self.imdb_job)

    @functools.cached_property
    def tmdb_job(self):
        """:class:`~.jobs.webdb.WebDbSearchJob` instance"""
        return jobs.webdb.WebDbSearchJob(
            query=self.content_path,
            db=self.tmdb,
            callbacks={
                'output': self._handle_tmdb_id,
            },
            precondition=self.make_precondition('tmdb_job'),
            **self.common_job_args(),
        )

    def _handle_tmdb_id(self, tmdb_id):
        # Update other webdb queries with TMDb info
        self.tracker.attach_task(self._propagate_webdb_info(self.tmdb, tmdb_id))

        if self._poster_webdb is self.tmdb:
            self._start_poster_job(self.tmdb, tmdb_id)

    @property
    def tmdb_id(self):
        """TMDb ID if :attr:`tmdb_job` is finished or `None`"""
        return self._get_webdb_id_from_job(self.tmdb_job)

    @functools.cached_property
    def tvmaze_job(self):
        """:class:`~.jobs.webdb.WebDbSearchJob` instance"""
        return jobs.webdb.WebDbSearchJob(
            query=self.content_path,
            db=self.tvmaze,
            callbacks={
                'output': self._handle_tvmaze_id,
            },
            precondition=self.make_precondition('tvmaze_job'),
            **self.common_job_args(),
        )

    def _handle_tvmaze_id(self, tvmaze_id):
        # Update other webdb queries with TVmaze info
        self.tracker.attach_task(self._propagate_webdb_info(self.tvmaze, tvmaze_id))

        if self._poster_webdb is self.tvmaze:
            self._start_poster_job(self.tvmaze, tvmaze_id)

    @property
    def tvmaze_id(self):
        """TVmaze ID if :attr:`tvmaze_job` is finished or `None`"""
        return self._get_webdb_id_from_job(self.tvmaze_job)

    def _get_webdb_id_from_job(self, webdb_job):
        if webdb_job.is_finished and webdb_job.output:
            return webdb_job.output[0]

    async def _propagate_webdb_info(self, webdb, webdb_id):
        target_webdb_jobs = [
            j for j in (getattr(self, f'{name}_job') for name in utils.webdbs.webdb_names())
            if (
                webdb.name not in j.name
                and j.is_enabled
                and not j.is_finished
            )
        ]

        title_english = await webdb.title_english(webdb_id)
        title_original = await webdb.title_original(webdb_id)
        query = utils.webdbs.Query(
            type=await webdb.type(webdb_id),
            title=title_english or title_original,
            year=await webdb.year(webdb_id),
        )

        _log.debug('Propagating %s info to: %r: %s', webdb.name, [j.name for j in target_webdb_jobs], query)
        for job in target_webdb_jobs:
            job.query.update(query)

        await self.update_release_name_from(webdb, webdb_id)

    @functools.cached_property
    def screenshots_job(self):
        """
        :class:`~.jobs.screenshots.ScreenshotsJob` instance

        The number of screenshots to make is taken the :attr:`screenshots_count`
        attribute.
        """
        return jobs.screenshots.ScreenshotsJob(
            content_path=self.content_path,
            exclude_files=self.exclude_files,
            count=self.screenshots_count,
            from_all_videos=self.screenshots_from_all_videos,
            optimize=self._screenshots_optimization,
            precondition=self.make_precondition('screenshots_job'),
            **self.common_job_args(),
        )

    @property
    def screenshots_count(self):
        """
        How many screenshots to make

        The default implementation uses :attr:`options`\\ ``["screenshots"]``
        with `None` as the default value, which creates a default number of
        screenshots.
        """
        return self.options.get('screenshots')

    @property
    def screenshots_from_all_videos(self):
        """
        Whether to make :attr:`screenshots_count` screenshots from all
        videos or just the first one

        See :meth:`.ScreenshotsJob.initialize`.

        The default implementation is always `False`.
        """
        return False

    image_host_config = {}
    """
    Dictionary that maps an image hosting service :attr:`~.ImageHostBase.name`
    to :attr:`~.ImageHostBase.default_config` values

    ``common`` is a special image host whose values are always applied.

    Values from a specific image hosting service overload ``common`` values.

    Example:

    >>> image_host_config = {
    ...     # Always generate 300p thumbnails
    ...     "common": {"thumb_width": 300},
    ...     # Use API key for specific image hosting service
    ...     "myhost": {"apikey": "d34db33f"},
    ... }
    """

    @functools.cached_property
    def upload_screenshots_job(self):
        """:class:`~.jobs.imghost.ImageHostJob` instance"""
        if self.image_hosts and self.screenshots_job:
            imghost_job = jobs.imghost.ImageHostJob(
                imghosts=self.image_hosts,
                precondition=self.make_precondition('upload_screenshots_job'),
                **self.common_job_args(),
            )
            # Timestamps and number of screenshots are determined in a
            # subprocess, we have to wait for that before we can set the number
            # of expected screenhots.
            self.screenshots_job.signal.register('screenshots_total', imghost_job.set_images_total)
            # Pass ScreenshotsJob's output to ImageHostJob input.
            self.screenshots_job.signal.register('output', imghost_job.enqueue)
            # Tell imghost_job to finish the current upload and then finish.
            self.screenshots_job.signal.register('finished', self.finalize_upload_screenshots_job)
            return imghost_job

    def finalize_upload_screenshots_job(self, _):
        self.upload_screenshots_job.finalize()

    @functools.cached_property
    def poster_job(self):
        """
        :class:`~.jobs.poster.PosterJob` instance

        This is started automatically when an appropriate webdb job finishes.
        """
        return jobs.poster.PosterJob(
            autostart=False,
            webdb_id=None,
            webdb=None,
            imghost=(
                self.image_hosts[0]
                if self.image_hosts else
                ()
            ),
            width=self.poster_max_width,
            height=self.poster_max_height,
            precondition=self.make_precondition('poster_job'),
            **self.common_job_args(),
        )

    def _start_poster_job(self, webdb, webdb_id):
        self.poster_job.start(
            webdb=webdb,
            webdb_id=webdb_id,
            season=self.release_name.only_season,
        )

    poster_max_width = 300
    """Maximum poster image width"""

    poster_max_height = 500
    """Maximum poster image height"""

    @functools.cached_property
    def _poster_webdb(self):
        """:class:`~.webdbs.base.WebDbApiBase` instance for finding poster image"""
        if (
                self.poster_job.is_enabled
                and not self.poster_job.is_started
                and self.poster_job in self.jobs_before_upload
        ):
            if (
                    self.tvmaze_job.is_enabled
                    and self.tvmaze_job in self.jobs_before_upload
                    and self.release_name.type in (utils.release.ReleaseType.season,
                                                   utils.release.ReleaseType.episode)
            ):
                return self.tvmaze

            elif (
                    self.imdb_job.is_enabled
                    and self.imdb_job in self.jobs_before_upload
            ):
                return self.imdb

            elif (
                    self.tmdb_job.is_enabled
                    and self.tmdb_job in self.jobs_before_upload
            ):
                return self.tmdb

            else:
                raise RuntimeError('You must add a WebDbSearchJob instance to jobs_before_upload')

    @functools.cached_property
    def mediainfo_job(self):
        """:class:`~.jobs.mediainfo.MediainfoJob` instance"""
        return jobs.mediainfo.MediainfoJob(
            content_path=self.content_path,
            from_all_videos=self.mediainfo_from_all_videos,
            exclude_files=self.exclude_files,
            precondition=self.make_precondition('mediainfo_job'),
            **self.common_job_args(),
        )

    @property
    def mediainfo_from_all_videos(self):
        """
        Whether to get the ``mediainfo`` output from all videos or just the
        first one

        See :meth:`.MediainfoJob.initialize`.

        The default implementation is always `False`.
        """
        return False

    @property
    def mediainfos_and_screenshots(self):
        """
        Map video file paths to mediainfo and screenshot URLs

        This property uses the results of :attr:`mediainfo_job` and
        :attr:`upload_screenshots_job`, and therefore both jobs must be
        :attr:`finished <upsies.jobs.base.JobBase.is_finished>`.

        Every key in the returned :class:`dict` is a path to a video file of
        which ``mediainfo`` was collected and/or screenshots were uploaded.

        Every value is a :class:`dict` with the keys ``mediainfo`` and
        ``screenshot_urls``.

        ``mediainfo`` is the return value of :func:`~.video.mediainfo` for the
        corresponding video file.

        ``screenshot_urls`` is a :class:`list` of screenshot URLs for the
        corresponding video file.

        .. note:: For DVDs (directories with a VIDEO_TS subdirectory),
                  ``mediainfo`` includes an .IFO and a .VOB key while
                  ``screenshot_urls`` is empty for the .IFO key.

                  ``mediainfo`` may also be an empty :class:`str`, for example
                  if :attr:`mediainfo_from_all_videos` is `False`.

        Season example:

        .. code::

            {
              "/path/to/season/Foo.S01E01.mkv": {
                "mediainfo": "<mediainfo for Foo.S01E01.mkv>",
                "screenshot_urls": [
                  "https://image.host/Foo.S01E01.mkv.0:48:08.png",
                  "https://image.host/Foo.S01E01.mkv.1:12:12.png",
                ]
              },
            }

        DVD example:

        .. code::

            {
              "/path/to/DVD/VIDEO_TS/VTS_01_0.IFO": {
                "mediainfo": "<mediainfo for VTS_01_0.IFO>",
                "screenshot_urls": [],
              }
              "/path/to/DVD/VIDEO_TS/VTS_01_1.VOB": {
                "mediainfo": "<mediainfo for VTS_01_1.VOB>",
                "screenshot_urls": [
                  "https://image.host/VTS_01_1.VOB.0:14:58.png",
                  "https://image.host/VTS_01_1.VOB.0:26:12.png"
                ]
              }
            }
        """
        mediainfos_and_screenshots = collections.defaultdict(
            lambda: {
                'mediainfo': '',
                'screenshot_urls': [],
            }
        )

        # NOTE: Don't expect all mediainfos/screenshots to exist. A fatal
        #       KeyError traceback is always worse than manually adding missing
        #       metadata to a semi-successful upload.

        assert self.mediainfo_job.is_finished
        for video_path, mediainfo in self.mediainfo_job.mediainfos_by_file.items():
            mediainfos_and_screenshots[video_path]['mediainfo'] = mediainfo

        # `screenshots_count` screenshots for each `video_path`
        assert self.upload_screenshots_job.is_finished
        urls_by_file = self.upload_screenshots_job.urls_by_file
        screenshots_by_file = self.screenshots_job.screenshots_by_file
        for video_path, screenshot_paths in screenshots_by_file.items():
            for screenshot_path in screenshot_paths:
                screenshot_url = urls_by_file.get(screenshot_path, None)
                if screenshot_url:
                    mediainfos_and_screenshots[video_path]['screenshot_urls'].append(screenshot_url)

        return dict(mediainfos_and_screenshots)

    @functools.cached_property
    def scene_check_job(self):
        """:class:`~.jobs.scene.SceneCheckJob` instance"""
        common_job_args = self.common_job_args(ignore_cache=True)
        common_job_args['force'] = self.options.get('is_scene')
        return jobs.scene.SceneCheckJob(
            content_path=self.content_path,
            precondition=self.make_precondition('scene_check_job'),
            **common_job_args,
        )

    def make_precondition(self, job_attr, precondition=None):
        """
        Return :attr:`~.base.JobBase.precondition` function for job

        The returned function takes into account :attr:`jobs_before_upload`,
        :attr:`jobs_after_upload` and :attr:`isolated_jobs`.

        :param str job_attr: Name of the job attribute this precondition is for

            By convention, this should be ``"<name>_job"``.

        :param callable precondition: Custom :attr:`~.base.JobBase.precondition`

            `precondition` must be either `None` or return anything truthy for
            the job to get enabled.
        """
        def custom_precondition(precondition=precondition):
            return precondition is None or precondition()

        def precondition():
            job = getattr(self, job_attr)
            if not (
                    job in self.jobs_before_upload
                    or job in self.jobs_after_upload
            ):
                # Subclass doesn't use this job
                return False

            isolated_jobs = self.isolated_jobs
            if isolated_jobs and job in isolated_jobs:
                # Jobs was isolated by user (i.e. other jobs are disabled)
                return custom_precondition()

            if not isolated_jobs:
                # No isolated jobs means all jobs in jobs_before/after_upload are enabled
                return custom_precondition()

            return False

        # Rename precondition function to make debugging more readable
        precondition.__qualname__ = f'{job_attr}_precondition'
        return precondition

    _NO_DEFAULT = object()

    def get_job_output(self, job, slice=None, default=_NO_DEFAULT):
        """
        Helper method for getting output from job

        `job` must be finished.

        :param job: :class:`~.jobs.base.JobBase` instance
        :param slice: :class:`int` to get a specific item from `job`'s output,
            `None` to return all output as a list, or a :class:`slice` object to
            return only one or more items of the output
        :param default: Default value if `job` is not finished or getting
            `slice` from `job`'s output fails.

        :raise RuntimeError: if `job` is not finished or getting `slice` from
            :attr:`~.base.JobBase.output` raises an :class:`IndexError`
        :return: :class:`list` or :class:`str`
        """
        if not job.is_finished:
            if default is not self._NO_DEFAULT:
                return default
            else:
                raise RuntimeError(f'Unfinished job: {job.name}')
        else:
            if slice is None:
                slice = builtins.slice(None, None)
            try:
                return job.output[slice]
            except IndexError:
                raise RuntimeError(f'Job finished with insufficient output: {job.name}: {job.output}')

    def get_job_attribute(self, job, attribute, default=_NO_DEFAULT):
        """
        Helper method for getting an attribute from job

        :param job: :class:`~.jobs.base.JobBase` instance
        :param str attribute: Name of attribute to get from `job`
        :param default: Default value if `job` is not finished

        :raise RuntimeError: if `job` is not finished
        :raise AttributeError: if `attribute` is not an attribute of `job`
        """
        if not job.is_finished:
            if default is not self._NO_DEFAULT:
                return default
            else:
                raise RuntimeError(f'Unfinished job: {job.name}')
        else:
            return getattr(job, attribute)

    def get_relative_file_path(self, filepath):
        """
        Return `filepath` relative to :attr:`content_path`

        The first path component of the returned path is the last component of
        :attr:`content_path`.

        :raise ValueError: if logical `filepath` is not a subpath of logical
            :attr:`content_path` ("logical" means that relative path components
            are resolved while symbolic links are not resolved to get an
            absolute path)
        """
        filepath_abs = pathlib.Path(filepath).absolute()
        content_path_parent = pathlib.Path(self.content_path).absolute().parent
        try:
            return str(filepath_abs.relative_to(content_path_parent))
        except ValueError:
            raise ValueError(f'{str(filepath_abs)!r} is not a subpath of {str(content_path_parent)!r}')

    def make_screenshots_grid(self, screenshots, columns=2, horizontal_spacer='  ', vertical_spacer='\n'):
        """
        Return BBcode for screenshots in a grid layout

        :param screenshots: Sequence of :class:`~.imghosts.common.UploadedImage`
            objects (URLs with a ``thumbnail_url`` attribute)
        :param int columns: How many columns to split screenshots into
        :param str horizontal_spacer: String between columns
        :param str vertical_spacer: String between rows

        :raise RuntimeError: if any screenshot doesn't have a thumbnail
        """
        groups = utils.as_groups(
            screenshots,
            group_sizes=(columns,),
            default=None,
        )

        rows = []
        for screenshots in groups:
            cells = []
            for screenshot in screenshots:
                # `screenshot` is `None` at the end if the number of screenshots
                # is not perfectly divisible by `columns`
                if screenshot is not None:
                    if screenshot.thumbnail_url is None:
                        raise RuntimeError(f'No thumbnail for {screenshot}')
                    cells.append(f'[url={screenshot}][img]{screenshot.thumbnail_url}[/img][/url]')

            # Space between columns
            rows.append(horizontal_spacer.join(cells))

        # Space between rows
        return vertical_spacer.join(rows)

    def read_nfo(self):
        """
        Find ``*.nfo`` file beneath :attr:`content_path` and return its contents

        If no ``*.nfo`` file is found, return `None`.

        See :func:`.string.read_nfo` for more information.
        """
        try:
            return utils.string.read_nfo(self.content_path)
        except errors.ContentError as e:
            self.error(e)



"""
Upload images to image hosting services
"""

from .. import errors, utils
from ..utils.imghosts import ImageHostBase
from . import QueueJobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImageHostJob(QueueJobBase):
    """Upload images to an image hosting service"""

    name = 'imghost'
    label = 'Image URLs'

    # Don't cache output and rely on caching in ImageHostBase. Otherwise, a
    # single failed/cancelled upload would throw away all the gathered URLs
    # because nothing is cached if a job fails.
    cache_id = None

    def initialize(self, *, imghosts, images_total=0, enqueue=()):
        """
        Validate arguments and set internal state

        :param imghosts: Sequence of :class:`ImageHostBase` subclass instances
            (see :func:`.utils.imghosts.imghost`)
        :param enqueue: Sequence of image paths (see
            :class:`~.base.QueueJobBase`)
        :param images_total: Number of images that are going to be uploaded. The
            only purpose of this value is to provide it via the :attr:`images_total`
            property to calculate progress.

        If `enqueue` is given, the job finishes after all images are uploaded.

        If `enqueue` is not given, calls to :meth:`upload` are expected and
        :meth:`finalize` must be called after the last call.

        `enqueue` and `images_total` must not be given at the same time.
        """
        if enqueue and images_total:
            raise RuntimeError('You must not give both arguments "enqueue" and "images_total".')
        else:
            for imghost in imghosts:
                assert isinstance(imghost, ImageHostBase), f'Not an ImageHostBase: {imghost!r}'
                # Force image hosts to cache image URLs in our cache directory
                imghost.cache_directory = self.cache_directory

            self._imghosts = imghosts
            self._uploaded_images = []
            self._urls_by_file = {}
            if images_total > 0:
                self.images_total = images_total
            else:
                self.images_total = len(enqueue)

    async def handle_input(self, image_path):
        if len(self._imghosts) <= 1:
            await self._upload_to_one(image_path)
        else:
            await self._upload_to_any(image_path)

    async def _upload_to_one(self, image_path):
        # Upload to 1 or 0 image hosts and error out immediately if that fails.
        for imghost in self._imghosts:
            try:
                await self._upload(image_path, imghost)
            except errors.RequestError as e:
                self.error(f'{imghost.name}: Upload failed: {utils.fs.basename(image_path)}: {e}')

    async def _upload_to_any(self, image_path):
        # Try each image host and stop on first successful upload.
        # Only warn about upload failures.
        fail = False
        for imghost in self._imghosts:
            try:
                await self._upload(image_path, imghost)
            except errors.RequestError as e:
                _log.debug('Failed to upload %s to %s: %r', image_path, imghost.name, e)
                self.warn(f'{imghost.name}: Upload failed: {utils.fs.basename(image_path)}: {e}')
                fail = True
            else:
                fail = False
                break
        if fail:
            self.error('All upload attempts failed.')

    async def _upload(self, image_path, imghost):
        info = await imghost.upload(image_path, cache=not self.ignore_cache)
        _log.debug('Uploaded image: %r', info)
        self._uploaded_images.append(info)
        self._urls_by_file[image_path] = info
        image_url = str(info)
        self.send(image_url)

    @property
    def exit_code(self):
        """`0` if all images were uploaded, `1` otherwise, `None` if unfinished"""
        if self.is_finished:
            if self.images_uploaded > 0 and self.images_uploaded == self.images_total:
                return 0
            else:
                return 1

    @property
    def uploaded_images(self):
        """
        Sequence of :class:`~.imghosts.common.UploadedImage` objects

        Use this property to get additional information like thumbnail URLs that
        are not part of this job's :attr:`~.base.JobBase.output`.
        """
        return tuple(self._uploaded_images)

    @property
    def urls_by_file(self):
        """Mapping of image file paths to image URLs"""
        return self._urls_by_file.copy()

    @property
    def images_uploaded(self):
        """Number of uploaded images"""
        return len(self._uploaded_images)

    @property
    def images_total(self):
        """Expected number of images to upload"""
        return self._images_total

    @images_total.setter
    def images_total(self, value):
        self._images_total = int(value)

    def set_images_total(self, value):
        """:attr:`images_total` setter as a method"""
        self.images_total = value



import functools

from . import constants, trackers, utils

defaults = {
    'config': {
        'main': {
            'cache_directory': utils.configfiles.config_value(
                value=constants.DEFAULT_CACHE_DIRECTORY,
                description='Where to store generated files.',
            ),
            'max_cache_size': utils.configfiles.config_value(
                value=utils.types.Bytes.from_string('20 MB'),
                description=(
                    'Maximum size of cache directory. '
                    'Units like "kB" and "MiB" are interpreted.'
                ),
            ),
        },
        'torrent-create': {
            'reuse_torrent_paths': utils.configfiles.config_value(
                value=[],
                description=(
                    'List of directories to search for a *.torrent file '
                    'to get piece hashes from instead of generating the '
                    'pieces hashes from file contents.'
                ),
            ),
        },
        'screenshots': {
            'optimize': utils.configfiles.config_value(
                value=utils.types.Choice('default', options=utils.image.optimization_levels),
                description='Whether to optimize screenshot file sizes',
            ),
        },
    },

    'trackers': {
        tracker.name: tracker.TrackerConfig()
        for tracker in trackers.trackers()
    },

    'imghosts': {
        imghost.name: imghost.default_config
        for imghost in utils.imghosts.imghosts()
    },

    'clients': {
        name: utils.btclient.client_defaults(name)
        for name in utils.btclient.client_names()
    },
}
"""Defaults for configuration options"""


@functools.lru_cache(maxsize=None)
def option_paths():
    """Tuple of configuration option paths (`<section>.<subsection>.<option>`)"""
    return utils.configfiles.ConfigFiles(defaults).paths


def option_type(option_path):
    filename, section, option = option_path.split('.', maxsplit=2)
    return type(defaults[filename][section][option])



from itertools import zip_longest

import pytest

from upsies.utils.types import ReleaseType
from upsies.utils.webdbs import Query, SearchResult, tmdb


@pytest.fixture
def api():
    return tmdb.TmdbApi()


@pytest.mark.parametrize('category', ('movie', 'tv'))
@pytest.mark.parametrize(
    argnames='text, exp_id',
    argvalues=(
        ('Some Title', None),
        ('1984', None),
        ('foo/123', None),
        ('{category}/123', '{category}/123'),
        ('https://www.themoviedb.org/{category}/123', '{category}/123'),
        ('https://www.themoviedb.org/{category}/123-foo-bar-baz', '{category}/123'),
        ('junk {category}/123 more junk', '{category}/123'),
        ('junk https://www.themoviedb.org/{category}/123-foo-bar-baz more junk', '{category}/123'),
    ),
    ids=lambda value: str(value),
)
def test_get_id_from_text(text, category, exp_id, api, store_response):
    text = text.format(category=category)
    if exp_id:
        exp_id = exp_id.format(category=category)
    return_value = api.get_id_from_text(text)
    assert return_value == exp_id


@pytest.mark.asyncio
async def test_search_handles_id_in_query(api, store_response):
    results = await api.search(Query(id='movie/525'))
    assert len(results) == 1

    assert results[0].id == 'movie/525'
    assert results[0].title == 'The Blues Brothers'
    assert results[0].type == ReleaseType.movie
    assert results[0].url == 'http://themoviedb.org/movie/525'
    assert results[0].year == '1980'

    # Make sure we're not awaiting the same coroutine again on subsequent calls
    for _ in range(2):
        assert (await results[0].cast())[:3] == ('Dan Aykroyd', 'John Belushi', 'James Brown')
        assert (await results[0].countries()) == ()
        assert (await results[0].directors()) == ('John Landis',)
        assert (await results[0].genres()) == ('music', 'comedy', 'action', 'crime')
        assert (await results[0].summary()) == (
            'Jake Blues, just released from prison, puts his old band back together '
            'to save the Catholic home where he and his brother Elwood were raised.'
        )
        assert (await results[0].title_english()) == ''
        assert (await results[0].title_original()) == 'The Blues Brothers'


@pytest.mark.asyncio
async def test_search_handles_non_unique_id_in_query(api, store_response):
    results = await api.search(Query(id='525'))
    assert len(results) == 2
    assert results[0].id == 'movie/525'
    assert results[1].id == 'tv/525'


@pytest.mark.asyncio
async def test_search_returns_empty_list_if_title_is_empty(api, store_response):
    assert await api.search(Query('', year='2009')) == []

@pytest.mark.asyncio
async def test_search_returns_list_of_SearchResults(api, store_response):
    results = await api.search(Query('Star Wars'))
    for result in results:
        assert isinstance(result, SearchResult)


@pytest.mark.parametrize(
    argnames=('query', 'exp_result'),
    argvalues=(
        (Query('alien', year='1979'), {'title': 'Alien', 'year': '1979'}),
        (Query('alien', year='1986'), {'title': 'Aliens', 'year': '1986'}),
        (Query('january', year='2012'), {'title': 'One Day in January', 'year': '2012'}),
        (Query('january', year='2015'), {'title': 'January Hymn', 'year': '2015'}),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_year(query, exp_result, api, store_response):
    results = await api.search(query)
    results_title_year = [
        {'title': r.title, 'year': r.year}
        for r in results
    ]
    assert exp_result in results_title_year


@pytest.mark.parametrize(
    argnames=('query', 'exp_titles'),
    argvalues=(
        (Query('Ash vs Evil Dead', type=ReleaseType.series), ('Ash vs Evil Dead',)),
        (Query('Ash vs Evil Dead', type=ReleaseType.movie), ()),
        (Query('Deadwood', type=ReleaseType.movie), ('Deadwood: The Movie',)),
        (Query('Deadwood', type=ReleaseType.series), ('Deadwood',)),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_type(query, exp_titles, api, store_response):
    results = await api.search(query)
    titles = {r.title for r in results}
    if exp_titles:
        for exp_title in exp_titles:
            assert exp_title in titles
    else:
        assert not titles


@pytest.mark.parametrize(
    argnames=('query', 'exp_cast'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), ('Dan Aykroyd', 'John Belushi')),
        (Query('February', year=2017), ('Ana Sofrenović', 'Aleksandar Đurica', 'Sonja Kolačarić')),
        (Query('Deadwood', year=2004), ('Timothy Olyphant', 'Ian McShane')),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_cast(query, exp_cast, api, store_response):
    results = await api.search(query)
    result = [r for r in results if r.title == query.title][0]
    cast = await result.cast()
    for member in exp_cast:
        assert member in cast


@pytest.mark.asyncio
async def test_search_result_countries(api, store_response):
    results = await api.search(Query('Star Wars'))
    for result in results:
        countries = await result.countries()
        assert countries == ()


@pytest.mark.parametrize(
    argnames=('query', 'exp_id'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), 'movie/525'),
        (Query('February', year=2017), 'movie/700711'),
        (Query('Deadwood', year=2004), 'tv/1406'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_id(query, exp_id, api, store_response):
    results = await api.search(query)
    result = [r for r in results if r.title == query.title][0]
    assert result.id == exp_id


@pytest.mark.parametrize(
    argnames=('query', 'exp_directors'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), ('John Landis',)),
        (Query('Deadwood', year=2004), ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_directors(query, exp_directors, api, store_response):
    results = await api.search(query)
    assert await results[0].directors() == exp_directors


@pytest.mark.parametrize(
    argnames=('query', 'exp_genres'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), ('music', 'comedy', 'action', 'crime')),
        (Query('February', year=2017), ('drama',)),
        (Query('Deadwood', year=2004), ('crime', 'western')),
        (Query('Farang', year=2017), ('drama',)),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_genres(query, exp_genres, api, store_response):
    results = await api.search(query)
    result = [r for r in results if r.title == query.title][0]
    genres = await result.genres()
    if exp_genres:
        for kw in exp_genres:
            assert kw in genres
    else:
        assert not genres


@pytest.mark.parametrize(
    argnames=('query', 'exp_summary'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), 'released from prison'),
        (Query('February', year='2017'), 'Husband and wife are going to their vacation cottage to overcome marriage crises'),
        (Query('Deadwood', year=2004), 'woven around actual historic events'),
        (Query('Lost & Found Music Studios', year=2015), 'singers-songwriters in an elite music program form bonds of friendship'),
        (Query('January', year=1973), ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_summary(query, exp_summary, api, store_response):
    results = await api.search(query)
    result = [r for r in results if r.title == query.title][0]
    summary = await result.summary()
    if exp_summary:
        assert exp_summary in summary
    else:
        assert summary == ''


@pytest.mark.parametrize(
    argnames=('query', 'exp_title'),
    argvalues=(
        (Query('Blues Brothers', year=1980), 'The Blues Brothers'),
        (Query('February', year=2017), "The Blackcoat's Daughter"),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_title(query, exp_title, api, store_response):
    results = await api.search(query)
    titles = [r.title for r in results]
    assert exp_title in titles


@pytest.mark.parametrize(
    argnames=('query', 'exp_title_english'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), ''),
        (Query('Deadwood', year=2004), ''),
        (Query('Le dolci signore', year=1968), 'Anyone Can Play'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_title_english(query, exp_title_english, api, store_response):
    results = await api.search(query)
    assert await results[0].title_english() == exp_title_english


@pytest.mark.parametrize(
    argnames=('query', 'exp_title_original'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), 'The Blues Brothers'),
        (Query('Deadwind', year=2018), 'Karppi'),
        (Query('Anyone Can Play', year=1968), 'Le dolci signore'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_title_original(query, exp_title_original, api, store_response):
    results = await api.search(query)
    assert await results[0].title_original() == exp_title_original


@pytest.mark.parametrize(
    argnames=('query', 'exp_type'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), ReleaseType.movie),
        (Query('February', year=2017), ReleaseType.movie),
        (Query('Deadwood', year=2004), ReleaseType.series),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_type(query, exp_type, api, store_response):
    results = await api.search(query)
    results_dict = {r.title: r for r in results}
    assert results_dict[query.title].type == exp_type


@pytest.mark.parametrize(
    argnames=('query', 'exp_url'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), 'http://themoviedb.org/movie/525'),
        (Query('February', year=2017), 'http://themoviedb.org/movie/700711'),
        (Query('Deadwood', year=2004), 'http://themoviedb.org/tv/1406'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_url(query, exp_url, api, store_response):
    results = await api.search(query)
    results_urls = [r.url for r in results]
    assert exp_url in results_urls


@pytest.mark.parametrize(
    argnames=('query', 'exp_title', 'exp_year'),
    argvalues=(
        (Query('Blues Brothers', year=1980), 'The Blues Brothers', '1980'),
        (Query('February', year='2017'), "The Blackcoat's Daughter", '2017'),
        (Query('Deadwood', year=2004), 'Deadwood', '2004'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_year(query, exp_title, exp_year, api, store_response):
    results = await api.search(query)
    results_dict = {r.title: r for r in results}
    assert results_dict[exp_title].year == exp_year

@pytest.mark.asyncio
async def test_search_result_parser_failure(api):
    result = tmdb._TmdbSearchResult(tmdb_api=api)
    assert result.id == ''
    assert result.title == ''
    assert result.type == ReleaseType.unknown
    assert result.url == ''
    assert result.year == ''
    assert await result.cast() == ()
    assert await result.countries() == ()
    assert await result.directors() == ()
    assert await result.genres() == ()
    assert await result.summary() == ''
    assert await result.title_english() == ''
    assert await result.title_original() == ''


@pytest.mark.parametrize(
    argnames=('id', 'exp_cast'),
    argvalues=(
        ('movie/525', (('Dan Aykroyd', 'http://themoviedb.org/person/707-dan-aykroyd'),
                       ('John Belushi', 'http://themoviedb.org/person/7171-john-belushi'))),
        ('movie/334536', (('Emma Roberts', 'http://themoviedb.org/person/34847-emma-roberts'),
                          ('Kiernan Shipka', 'http://themoviedb.org/person/934289-kiernan-shipka'))),
        ('tv/1406', (('Timothy Olyphant', 'http://themoviedb.org/person/18082-timothy-olyphant'),
                     ('Ian McShane', 'http://themoviedb.org/person/6972-ian-mcshane'))),
        ('tv/74802', (('Pihla Viitala', 'http://themoviedb.org/person/93564-pihla-viitala'),
                      ('Lauri Tilkanen', 'http://themoviedb.org/person/124867-lauri-tilkanen'))),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_cast(id, exp_cast, api, store_response):
    cast = (await api.cast(id))[:2]
    if not cast:
        assert exp_cast == ()
    else:
        for person, (name, url) in zip_longest(cast, exp_cast):
            assert person == name
            assert person.url == url


@pytest.mark.parametrize(argnames='id', argvalues=('movie/525', 'movie/334536', 'tv/1406', 'tv/74802'))
@pytest.mark.asyncio
async def test_countries(id, api, store_response):
    countries = await api.countries(id)
    assert countries == ()


@pytest.mark.parametrize(
    argnames=('id', 'exp_creators'),
    argvalues=(
        ('movie/125244', ()),
        ('movie/334536', ()),
        ('tv/1406', (('David Milch', 'http://themoviedb.org/person/151295-david-milch'),)),
        ('tv/74802', (('Rike Jokela', 'http://themoviedb.org/person/140497-rike-jokela'),)),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_creators(id, exp_creators, api, store_response):
    creators = await api.creators(id)
    if not creators:
        assert exp_creators == ()
    else:
        for person, (name, url) in zip_longest(creators, exp_creators):
            assert person == name
            assert person.url == url


@pytest.mark.parametrize(
    argnames=('id', 'exp_directors'),
    argvalues=(
        ('movie/125244', (('James Algar', 'http://themoviedb.org/person/5690-james-algar'),)),
        ('movie/334536', (('Oz Perkins', 'http://themoviedb.org/person/90609-oz-perkins'),)),
        ('tv/1406', ()),
        ('tv/74802', ()),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_directors(id, exp_directors, api, store_response):
    directors = await api.directors(id)
    if not directors:
        assert exp_directors == ()
    else:
        for person, (name, url) in zip_longest(directors, exp_directors):
            assert person == name
            assert person.url == url


@pytest.mark.parametrize(
    argnames=('id', 'exp_genres'),
    argvalues=(
        ('movie/525', ('music', 'comedy', 'action', 'crime')),
        ('movie/334536', ('horror', 'thriller')),
        ('tv/1406', ('crime', 'western')),
        ('tv/74802', ('drama', 'crime')),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_genres(id, exp_genres, api, store_response):
    genres = await api.genres(id)
    if exp_genres:
        for member in exp_genres:
            assert member in genres
    else:
        assert not genres


@pytest.mark.parametrize(
    argnames='id, exp_url',
    argvalues=(
        # The Blues Brothers
        ('movie/525', 'http://themoviedb.org/t/p/w300_and_h450_bestv2/3DiSrcYELCLkwnjl9EZp2pkKGep.jpg'),
        # The Blackcoat's Daughter
        ('movie/334536', 'http://themoviedb.org/t/p/w300_and_h450_bestv2/zvcdDfwxn0F3Rky3m3Dl2eSVX5X.jpg'),
        # Deadwood
        ('tv/1406', 'http://themoviedb.org/t/p/w300_and_h450_bestv2/4Yp35DVbVOAWkfQUIQ7pbh3u0aN.jpg'),
        # Karppi
        ('tv/74802', 'http://themoviedb.org/t/p/w300_and_h450_bestv2/cUKqWS2v7D6DVKQze2Iz2netwRH.jpg'),
        # Lost & Found Music Studios
        ('tv/66260', 'http://themoviedb.org/t/p/w300_and_h450_bestv2/2febzBrLWIQ9oAlZAvnm4XN0cu1.jpg'),
        # Le dolci signore
        ('movie/3405', 'http://themoviedb.org/t/p/w300_and_h450_bestv2/sf31ob8bHZsgCH01J8PutaOnaTA.jpg'),
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_poster_url(id, exp_url, api, store_response):
    poster_url = await api.poster_url(id)
    assert poster_url == exp_url


@pytest.mark.parametrize(
    argnames=('id', 'exp_runtimes'),
    argvalues=(
        # The Blues Brothers
        ('movie/525', {'default': 133}),
        # The Blackcoat's Daughter
        ('movie/334536', {'default': 93}),
        # Deadwood
        ('tv/1406', {}),
        # Karppi
        ('tv/74802', {}),
        # Lost & Found Music Studios
        ('tv/66260', {}),
        # Le dolci signore
        ('movie/3405', {'default': 88}),
        ('tv/111086', {}),  # no runtimes found
        (None, {}),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_runtimes(id, exp_runtimes, api, store_response):
    runtimes = await api.runtimes(id)
    assert runtimes == exp_runtimes


@pytest.mark.parametrize(
    argnames=('id', 'exp_rating'),
    argvalues=(
        # The Blues Brothers
        ('movie/525', 77.0),
        # The Blackcoat's Daughter
        ('movie/334536', 58.0),
        # Deadwood
        ('tv/1406', 81.0),
        # Karppi
        ('tv/74802', 69.0),
        # Lost & Found Music Studios
        ('tv/66260', 84.0),
        # Le dolci signore
        ('movie/3405', 18.0),
        (None, None),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_rating(id, exp_rating, api, store_response):
    rating = await api.rating(id)
    assert rating == exp_rating


@pytest.mark.parametrize(
    argnames=('id', 'exp_summary'),
    argvalues=(
        ('movie/525', 'released from prison'),
        ('movie/334536', 'Two young students at a prestigious prep school for girls'),
        ('tv/1406', 'woven around actual historic events'),
        ('tv/74802', 'the body of a young woman on a construction site'),
        ('tv/66260', ''),
        ('movie/3405', ''),
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_summary(id, exp_summary, api, store_response):
    summary = await api.summary(id)
    assert exp_summary in summary


@pytest.mark.parametrize(
    argnames=('id', 'exp_title_english', 'exp_title_original'),
    argvalues=(
        ('movie/221437', '', 'Miss ZOMBIE'),
        ('movie/11841', 'The 36th Chamber of Shaolin', '少林三十六房'),
        ('movie/334536', '', "The Blackcoat's Daughter"),
        ('movie/3405', 'Anyone Can Play', 'Le dolci signore'),
        ('movie/525', '', 'The Blues Brothers'),
        ('movie/22156', 'The Nest', 'Nid de guêpes'),
        ('tv/1406', '', 'Deadwood'),
        ('tv/66260', '', 'Lost & Found Music Studios'),
        ('tv/74802', 'Deadwind', 'Karppi'),
        (None, '', ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_title_english_original(id, exp_title_english, exp_title_original, api, store_response):
    assert await api.title_original(id) == exp_title_original
    assert await api.title_english(id) == exp_title_english
    assert await api.title_english(id, default_to_original=False) == exp_title_english
    if exp_title_english:
        assert await api.title_english(id, default_to_original=True) == exp_title_english
    else:
        assert await api.title_english(id, default_to_original=True) == exp_title_original


@pytest.mark.parametrize(
    argnames=('id', 'exp_type'),
    argvalues=(
        ('movie/525', ReleaseType.movie),
        ('tv/525', ReleaseType.series),
        ('tv/1406', ReleaseType.series),
        ('movie/1406', ReleaseType.movie),
        (None, ReleaseType.unknown),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_type(id, exp_type, api, store_response):
    type = await api.type(id)
    assert type is exp_type


@pytest.mark.parametrize(
    argnames=('id', 'exp_url'),
    argvalues=(
        ('movie/525', f'{tmdb.TmdbApi._url_base}/movie/525'),
        ('/tv/1406', f'{tmdb.TmdbApi._url_base}/tv/1406'),
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_url(id, exp_url, api):
    assert await api.url(id) == exp_url


@pytest.mark.parametrize(
    argnames=('id', 'exp_year'),
    argvalues=(
        ('movie/525', '1980'),
        ('movie/334536', '2017'),
        ('tv/1406', '2004'),
        ('tv/74802', '2018'),
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_year(id, exp_year, api, store_response):
    assert await api.year(id) == exp_year



"""
Get information from the user
"""

import collections
import inspect
import re

from .. import utils
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ChoiceJob(JobBase):
    """
    Ask the user to choose from a set of values

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``dialog_updated``
            Emitted when the :attr:`options`, :attr:`focused` or
            :attr:`autodetected` properties are set or when :attr:`options` is
            modified. Registered callbacks get the job instance as a positional
            argument.
    """

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    @property
    def options(self):
        """
        Sequence of options the user can make

        An option is a :class:`tuple` with 2 items. The first item is a
        human-readable :class:`str` and the second item is any object that is
        available as :attr:`choice` when the job is finished.

        Options may also be passed as a flat iterable of :class:`str`, in which
        case both items in the tuple are identical.

        When setting this property, focus is preserved if the value of the
        focused option exists in the new options. Otherwise, the first option is
        focused.
        """
        return getattr(self, '_options', ())

    @options.setter
    def options(self, options):
        # Build new list of options
        valid_options = []
        for option in options:
            if isinstance(option, str):
                valid_options.append((option, option))
            elif isinstance(option, collections.abc.Sequence):
                if len(option) != 2:
                    raise ValueError(f'Option must be 2-tuple, not {option!r}')
                else:
                    valid_options.append((str(option[0]), option[1]))
            else:
                raise ValueError(f'Option must be 2-tuple, not {option!r}')

        if len(valid_options) < 2:
            raise ValueError(f'There must be at least 2 options: {options!r}')

        # Remember current focus
        prev_focused = self.focused

        # Set new options
        self._options = utils.MonitoredList(
            valid_options,
            callback=lambda _: self.signal.emit('dialog_updated', self),
        )

        # Try to restore focus if the previously focused item still exists,
        # default to first option
        self._focused_index = 0
        if prev_focused:
            prev_label, prev_value = prev_focused
            for index, (label, value) in enumerate(valid_options):
                if value == prev_value:
                    self._focused_index = index
                    break

        self.signal.emit('dialog_updated', self)

    def get_index(self, thing):
        """
        Return index in :attr:`options`

        :param thing: Identifier of a option in :attr:`options`

            This can be:

            `None`
                Return `None`

            an index (:class:`int`)
               Return `thing`, but limited to the minimum/maximum valid index.

            one of the 2-tuples in :attr:`options`
                Return the index of `thing` in :attr:`options`.

            an item of one of the 2-tuples in :attr:`options`
                Return the index of the first 2-tuple in :attr:`options` that
                contains `thing`.

            a :func:`regular expression <re.compile>`
                Return the index of the first 2-tuple in :attr:`options` that
                contains something that matches `thing`. Non-string values are
                converted to :class:`str` for matching against the regular
                expression.

        :raise ValueError: if `thing` is not found in :attr:`options`
        """
        if thing is None:
            return None

        elif isinstance(thing, int):
            return max(0, min(thing, len(self.options) - 1))

        elif thing in self.options:
            return self.options.index(thing)

        else:
            for i, (label, value) in enumerate(self.options):
                # Focus by human-readable text or value
                if thing == label or thing == value:
                    return i

                # Focus by matching regex against human-readable text or value
                elif isinstance(thing, re.Pattern):
                    value_str = str(value)
                    if thing.search(label) or thing.search(value_str):
                        return i

        raise ValueError(f'No such option: {thing!r}')

    def get_option(self, thing):
        """
        Return item in :attr:`options`

        :param thing: See :meth:`get_index`

        If `thing` is `None`, return the currently focused option, which is
        indicated by :attr:`focused_index`.
        """
        option_index = self.get_index(thing)
        if option_index is None:
            option_index = self.focused_index
        return self.options[option_index]

    @property
    def focused_index(self):
        """Index of currently focused option in :attr:`options`"""
        return getattr(self, '_focused_index', None)

    @property
    def focused(self):
        """
        Currently focused option (2-tuple)

        This property can be set to anything that is a valid value for
        :meth:`get_index`.
        """
        focused_index = self.focused_index
        if focused_index is not None:
            return self.options[focused_index]

    @focused.setter
    def focused(self, focused):
        # focused_index can't be set to None, so we default to first option
        self._focused_index = self.get_index(focused)
        if self._focused_index is None:
            self._focused_index = 0
        self.signal.emit('dialog_updated', self)

    @property
    def autodetected_index(self):
        """Index of autodetected option in :attr:`options` or `None`"""
        return getattr(self, '_autodetected_index', None)

    @property
    def autodetected(self):
        """
        Autodetected option (2-tuple) or `None`

        This property can be set to anything that is a valid value for
        :meth:`get_index`.
        """
        autodetected_index = self.autodetected_index
        if autodetected_index is not None:
            return self.options[autodetected_index]

    @autodetected.setter
    def autodetected(self, autodetected):
        self._autodetected_index = self.get_index(autodetected)
        self.signal.emit('dialog_updated', self)

    @property
    def choice(self):
        """
        User-chosen value if job is finished, `None` otherwise

        While :attr:`~.base.JobBase.output` contains the user-readable string
        (first item of the chosen 2-tuple in :attr:`options`), this is the
        object attached to it (second item).
        """
        return getattr(self, '_choice', None)

    def _set_choice(self, choice):
        # This method is called via `output` signal (see initialize()), which is
        # emitted by send(), so make_choice() doesn't need to call _set_choice().
        if hasattr(self, '_choice'):
            raise RuntimeError(f'{self.name}: Choice was already made: {self.choice}')
        else:
            label, value = self.get_option(choice)
            self._choice = value

    def make_choice(self, thing):
        """
        Make a choice and :meth:`~.JobBase.finish` this job

        :param thing: See :meth:`get_option`

        After this method is called, this job :attr:`~.JobBase.is_finished`,
        :attr:`~.JobBase.output` contains the human-readable label of the choice
        and :attr:`choice` is the machine-readable value of the choice.
        """
        label, value = self.get_option(thing)
        self.send(label)
        self.finish()

    def set_label(self, identifier, new_label):
        """
        Assign new label to option

        :param identifier: Option (2-tuple of `(<current label>, <value>)`) or the
            current label or value of an option
        :param new_label: New label for the option defined by `identifier`

        Do nothing if `identifier` doesn't match any option.
        """
        new_options = []
        for label, value in tuple(self.options):
            if identifier == label or identifier == value or identifier == (label, value):
                new_options.append((str(new_label), value))
            else:
                new_options.append((label, value))
        if self.options != new_options:
            self.options = new_options

    def initialize(self, *, name, label, options, focused=None, autodetected=None,
                   autodetect=None, autodetect_and_finish=None):
        """
        Set internal state

        :param name: Name for internal use
        :param label: Name for user-facing use
        :param options: Iterable of options the user can pick from
        :param focused: See :attr:`focused`
        :param autodetected: See :attr:`autodetected`
        :param autodetect: Callable that sets :attr:`autodetected` in
            :meth:`.JobBase.start`

            `autodetect` gets the job instance (``self``) as a positional
            argument.
        :param autodetect_and_finish: Same as `autodetect`, but :meth:`finish`
            this job if the returned value is not `None`

        :raise ValueError: if `options` is shorter than 2 or `focused` is
            invalid
        """
        self._name = str(name)
        self._label = str(label)
        self._autodetect = autodetect
        self._autodetect_and_finish = autodetect_and_finish

        self.signal.add('dialog_updated')
        self.signal.register('output', self._set_choice)

        self.options = options
        self.autodetected = autodetected

        if focused is not None:
            self.focused = focused
        else:
            self.focused = autodetected

    def start(self):
        super().start()

        # Job is not started if it is disabled.
        # Job is finished right away if output was cached previously.
        if self.is_started and not self.is_finished:
            if self._autodetect:
                autodetected = self._autodetect(self)
                if autodetected is not None:
                    self.autodetected = self.focused = autodetected

            if self._autodetect_and_finish:
                autodetected = self._autodetect_and_finish(self)
                if autodetected is not None:
                    self.autodetected = self.focused = autodetected
                    self.make_choice(self.focused)


class TextFieldJob(JobBase):
    """
    Ask the user for text input

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``text``
            Emitted when :attr:`text` was changed without user input. Registered
            callbacks get the new text as a positional argument.

        ``is_loading``
            Emitted when :attr:`is_loading` was changed. Registered callbacks
            get the new :attr:`is_loading` value as a positional argument.

        ``read_only``
            Emitted when :attr:`read_only` was changed. Registered callbacks get
            the new :attr:`read_only` value as a positional argument.

        ``obscured``
            Emitted when :attr:`obscured` was changed. Registered callbacks get
            the new :attr:`obscured` value as a positional argument.
    """

    @property
    def name(self):
        return self._name

    @property
    def label(self):
        return self._label

    @property
    def text(self):
        """Current text"""
        return getattr(self, '_text', '')

    @text.setter
    def text(self, text):
        # Don't call validator here because it should be possible to set invalid
        # texts and let the user fix it manually. We only validate when we
        # commit to `text` (see send()).
        self._text = self._normalizer(str(text))
        self.signal.emit('text', self.text)

    @property
    def obscured(self):
        """
        Whether the text is unreadable, e.g. when entering passwords

        This is currently not fully implemented.
        """
        return getattr(self, '_obscured', False)

    @obscured.setter
    def obscured(self, obscured):
        self._obscured = bool(obscured)
        self.signal.emit('obscured', self.obscured)

    @property
    def read_only(self):
        """Whether the user can change the text"""
        return getattr(self, '_read_only', False)

    @read_only.setter
    def read_only(self, read_only):
        self._read_only = bool(read_only)
        self.signal.emit('read_only', self.read_only)

    @property
    def is_loading(self):
        """Whether :attr:`text` is currently being changed automatically"""
        return getattr(self, '_is_loading', False)

    @is_loading.setter
    def is_loading(self, is_loading):
        self._is_loading = bool(is_loading)
        self.signal.emit('is_loading', self.is_loading)

    def initialize(self, *, name, label,
                   text='', default=None, finish_on_success=False, nonfatal_exceptions=(),
                   validator=None, normalizer=None, obscured=False, read_only=False):
        """
        Set internal state

        :param name: Name for internal use
        :param label: Name for user-facing use
        :param text: Initial text or callable (synchronous or asynchronous)
            which will be called when the job is :meth:`executed
            <upsies.jobs.base.JobBase.execute>`. The return value is used as the
            initial text.
        :param default: Text to use if `text` is `None`, returns `None` or
            raises `nonfatal_exceptions`
        :param finish_on_success: Whether to call :meth:`finish` after setting
            :attr:`text` to `text`

            .. note:: :meth:`finish` is not called when :attr:`text` is set to
                      `default`.
        :param nonfatal_exceptions: Sequence of exception classes that may be
            raised by `text` without crashing

            Fatal exceptions are raised immediately while non-fatal exceptions
            are passed to :meth:`warn`, and the user can fix the situation
            manually.
        :param validator: Callable that gets text before job is finished. If
            `ValueError` is raised, it is displayed as a warning instead of
            finishing the job.
        :type validator: callable or None
        :param normalizer: Callable that gets text and returns the new text. It
            is called before `validator`. It should not raise any exceptions.
        :type normalizer: callable or None
        :param bool obscured: Whether :attr:`obscured` is set to `True`
            initially (currently fully implemented)
        :param bool read_only: Whether :attr:`read_only` is set to `True`
            initially
        """
        self._name = str(name)
        self._label = str(label)
        self._validator = validator or (lambda _: None)
        self._normalizer = normalizer or (lambda text: text)
        self.signal.add('text')
        self.signal.add('is_loading')
        self.signal.add('read_only')
        self.signal.add('obscured')
        self.obscured = obscured
        self.read_only = read_only

        if isinstance(text, str):
            # Set text attribute immediately
            self.text = text

        # Get text from callable when job is executed
        self._execute_arguments = (text, default, finish_on_success, nonfatal_exceptions)

    def execute(self):
        text, default, finish_on_success, nonfatal_exceptions = self._execute_arguments

        if inspect.isawaitable(text):
            self.attach_task(
                self.fetch_text(
                    coro=text,
                    default=default,
                    finish_on_success=finish_on_success,
                    nonfatal_exceptions=nonfatal_exceptions,
                )
            )

        elif inspect.iscoroutinefunction(text):
            self.attach_task(
                self.fetch_text(
                    coro=text(),
                    default=default,
                    finish_on_success=finish_on_success,
                    nonfatal_exceptions=nonfatal_exceptions,
                )
            )

        else:
            self.set_text(
                text=text,
                default=default,
                finish_on_success=finish_on_success,
                nonfatal_exceptions=nonfatal_exceptions,
            )

    def set_text(self, text, default=None, finish_on_success=False, nonfatal_exceptions=()):
        """
        Change :attr:`text` value

        :param text: New text value

            If this is callable, it must return the new :attr:`text` or `None`
            to use `default`

        :param default: Text to use if `text` is `None`, returns `None` or
            raises `nonfatal_exceptions`

        :param finish_on_success: Whether to call :meth:`finish` after setting
            :attr:`text` to `text`

            .. note:: :meth:`finish` is not called when :attr:`text` is set to
                      `default`.

        :param nonfatal_exceptions: Sequence of exception classes that may be
            raised by `text` without crashing

            Fatal exceptions are raised immediately while non-fatal exceptions
            are passed to :meth:`warn`, and the user can fix the situation
            manually.
        """
        self.is_loading = self.read_only = True
        try:
            if callable(text):
                new_text = text()
            elif text is not None:
                new_text = str(text)
            else:
                new_text = None
        except Exception as e:
            self._set_text(default)
            self._handle_exception(e, nonfatal=nonfatal_exceptions)
        else:
            self._set_text(new_text, default=default, finish=finish_on_success)
        finally:
            # Always re-enable text field after we're done messing with it
            self.is_loading = self.read_only = False

    async def fetch_text(self, coro, default=None, finish_on_success=False, nonfatal_exceptions=()):
        """
        Get :attr:`text` from coroutine

        :param coro: Coroutine that returns the new :attr:`text` or `None` to
            use `default`

        :param default: Text to use if `text` is `None`, returns `None` or
            raises `nonfatal_exceptions`

        :param finish_on_success: Whether to call :meth:`finish` after setting
            :attr:`text` to `coro` return value

            .. note:: :meth:`finish` is not called when :attr:`text` is set to
                      `default`.

        :param nonfatal_exceptions: Sequence of exception classes that may be
            raised by `coro` without crashing

            Fatal exceptions are raised immediately while non-fatal exceptions
            are passed to :meth:`warn`, and the user can fix the situation
            manually.
        """
        self.is_loading = self.read_only = True
        try:
            new_text = await coro
        except Exception as e:
            _log.debug('CAUGHT EXCEPTION: %r', e)
            self._set_text(default)
            self._handle_exception(e, nonfatal=nonfatal_exceptions)
        else:
            self._set_text(new_text, default=default, finish=finish_on_success)
        finally:
            # Always re-enable text field after we're done messing with it
            self.is_loading = self.read_only = False

    def _set_text(self, text, default=None, finish=False):
        # Fill in text
        if text is not None:
            self.text = text
            # Only finish if the intended text was set
            if finish:
                # We must call send() because it handles output caching, and it also
                # finishes the job.
                self.send(text)

        elif default is not None:
            self.text = default

    def _handle_exception(self, exception, nonfatal=()):
        # Display error and allow user to manually fix the situation
        if isinstance(exception, nonfatal):
            self.warn(exception)

        # An exception we didn't expect
        else:
            raise exception

    def send(self, output):
        """
        Validate `output` before actually sending it

        Pass `output` to `validator`. If :class:`ValueError` is raised, pass it
        to :meth:`warn` and do not finish. Otherwise, pass `output` to
        :meth:`~.base.JobBase.send` and :meth:`~.base.JobBase.finish` this job.
        """
        output = self._normalizer(output)
        try:
            self._validator(output)
        except ValueError as e:
            self.warn(e)
        else:
            super().send(output)
            super().finish()



"""
API for imdb.com
"""

import functools
import re

from ... import utils
from ..types import ReleaseType
from . import common
from .base import WebDbApiBase

import logging  # isort:skip
_log = logging.getLogger(__name__)


class ImdbApi(WebDbApiBase):
    """API for imdb.com"""

    name = 'imdb'
    label = 'IMDb'

    default_config = {}

    _url_base = 'https://www.imdb.com'
    _soup_cache = {}

    async def _get_soup(self, path, params={}):
        path = path.rstrip('/')

        cache_id = (path, tuple(sorted(params.items())))
        if cache_id in self._soup_cache:
            return self._soup_cache[cache_id]

        text = await utils.http.get(
            url=f'{self._url_base}/{path}/',
            params=params,
            user_agent=await utils.http.get_popular_user_agent(),
            cache=True,
        )

        self._soup_cache[cache_id] = utils.html.parse(text)
        return self._soup_cache[cache_id]

    _title_types = {
        ReleaseType.movie: 'feature,tv_movie,short,video,tv_short',
        ReleaseType.season: 'tv_series,tv_miniseries',
        # Searching for single episodes is currently not supported
        ReleaseType.episode: 'tv_series,tv_miniseries',
    }

    def sanitize_query(self, query):
        """
        Deal with IMDb-specific quirks

        - Remove ``"and"`` from :attr:`.Query.title` because IMDb doesn't find
          ``"Foo & Bar"`` if we search for ``"Foo and Bar"``. It's seems to work
          vice versa, i.e. the query ``"Foo and Bar"`` finds ``"Foo & Bar"``, so
          we keep any ``"&"``.

        - Replace ``"dont"`` with ``"don't"``, ``"cant"`` with ``"can't"``.
        """
        query = super().sanitize_query(query)
        query.title = re.sub(r'\s(?i:and)(\s)', r'\1', query.title)
        query.title = re.sub(r'\b(?i:dont)(\b)', "don't", query.title)
        query.title = re.sub(r'\b(?i:cant)(\b)', "can't", query.title)
        query.title = re.sub(r'\b(?i:wont)(\b)', "won't", query.title)
        return query

    def get_id_from_text(self, text):
        # Example: https://www.imdb.com/title/tt0048918/
        match = re.search(r'\b(tt\d+)\b', text)
        if match:
            return match.group(1)

    async def search(self, query):
        _log.debug('Searching IMDb for %s', query)

        if query.id:
            title_english = await self.title_english(query.id)
            title_original = await self.title_original(query.id)
            return [_ImdbSearchResult(
                imdb_api=self,
                cast=functools.partial(self.cast, query.id),
                countries=functools.partial(self.countries, query.id),
                directors=functools.partial(self.directors, query.id),
                genres=functools.partial(self.genres, query.id),
                id=query.id,
                summary=functools.partial(self.summary, query.id),
                title=title_english or title_original,
                title_english=title_english,
                title_original=title_original,
                type=await self.type(query.id),
                url=await self.url(query.id),
                year=await self.year(query.id),
            )]

        elif not query.title:
            return []

        else:
            path = 'search/title/'
            params = {'title': query.title_normalized if not query.id else query.id}
            if query.type is not ReleaseType.unknown:
                params['title_type'] = self._title_types[query.type]
            if query.year is not None:
                params['release_date'] = f'{query.year}-01-01,{query.year}-12-31'

            def is_search_result(tag):
                # Old website design
                if tag.get('class') == ['lister-item-content']:
                    header = tag.find(class_='lister-item-header')
                    if not header or 'Episode:' not in header.strings:
                        return True

                # New design
                if 'ipc-metadata-list-summary-item__tc' in tag.get('class', ()):
                    return True

                return False

            soup = await self._get_soup(path, params=params)
            items = soup.find_all(is_search_result)
            results = [
                _ImdbSearchResult(soup=item, imdb_api=self)
                for item in items
            ]
            return results

    _person_url_path_regex = re.compile(r'(/name/nm\d+)')

    def _get_persons(self, tag):
        a_tags = tag.find_all('a', href=self._person_url_path_regex)
        persons = []
        for a_tag in a_tags:
            if a_tag.string:
                name = a_tag.string.strip()
                url_path = self._person_url_path_regex.match(a_tag["href"]).group(1)
                url = f'{self._url_base.rstrip("/")}/{url_path.lstrip("/")}'
                persons.append(common.Person(name, url))
        return tuple(persons)

    async def cast(self, id):
        cast = []
        if id:
            soup = await self._get_soup(f'title/{id}/')
            # New website
            cast_tag = soup.find(class_='title-cast__grid')
            if cast_tag is None:
                # Old website
                cast_tag = soup.find(class_='cast_list')
            if cast_tag:
                cast.extend(self._get_persons(cast_tag))
        return tuple(cast)

    async def _countries(self, id):
        countries = []
        if id:
            soup = await self._get_soup(f'title/{id}/')
            a_tags = soup.find_all(href=re.compile(r'/search/title.*?country_of_origin='))
            for a_tag in a_tags:
                country = ''.join(a_tag.stripped_strings)
                countries.append(country)
        return tuple(countries)

    _creators_label_regex = re.compile('^Creators?:?$')

    async def creators(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/')
            # Old website design
            credits_tags = soup.find_all(class_='credit_summary_item')
            if not credits_tags:
                # New website design
                credits_tags = soup.find_all(class_='ipc-metadata-list__item')
            for tag in credits_tags:
                if any(self._creators_label_regex.search(string)
                       for string in tag.stripped_strings):
                    return self._get_persons(tag)
        return ()

    _directors_label_regex = re.compile('^Directors?:?$')

    async def directors(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/')
            # Old website design
            credits_tags = soup.find_all(class_='credit_summary_item')
            if not credits_tags:
                # New website design
                credits_tags = soup.find_all(class_='ipc-metadata-list__item')
            for tag in credits_tags:
                if any(self._directors_label_regex.search(string)
                       for string in tag.stripped_strings):
                    return self._get_persons(tag)
        return ()

    async def genres(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/')
            genres = []
            genre_links = soup.find_all('a', href=re.compile(r'/search/title\?.*genres='))
            for genre_link in genre_links:
                match = re.search(r'genres=([^&]+)', genre_link['href'])
                if match:
                    genre = match.group(1).lower()
                    if genre not in genres:
                        genres.append(genre)
            return tuple(genres)
        return ()

    async def poster_url(self, id, season=None):
        if id and await self.type(id) != ReleaseType.episode:  # noqa: E721: do not compare types
            soup = await self._get_soup(f'title/{id}/reference')
            tag = soup.select_one('meta [property="og:image"]')
            if tag:
                url = tag.get('content')
                if url:
                    if '@' in url:
                        # Remove any cropping instructions and other stuff
                        url = re.sub(r'_CR[\d+,]+', r'', url)
                    return url

        return ''

    rating_min = 0.0
    rating_max = 10.0

    async def rating(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/')

            # Old website design
            rating_tag = soup.find(itemprop='ratingValue')

            if not rating_tag:
                # New website design
                rating_tag = soup.find(attrs={'data-testid': 'hero-rating-bar__aggregate-rating__score'})
                if rating_tag and rating_tag.children:
                    rating_tag = rating_tag.contents[0]

            if rating_tag:
                try:
                    return float(rating_tag.string)
                except (ValueError, TypeError):
                    pass

        return None

    _ignored_runtimes_keys = (
        re.compile(r'^(?i:approx)\w*$'),
    )

    async def _runtimes(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/technical')
            label_tags = soup.select('.ipc-metadata-list__item > .ipc-metadata-list-item__label')
            for tag in label_tags:
                if 'Runtime' in utils.html.as_text(tag):
                    values_tag = tag.next_sibling
                    if values_tag:
                        return self._parse_runtimes(values_tag)
        return {}

    def _parse_runtimes(self, tag):
        runtimes = {}
        for li_tag in tag.find_all('li'):
            li_text = ' '.join(li_tag.stripped_strings)

            # Examples: 2h 13m (133 min)
            #           2h 28m (148 min) (extended) (United States)
            #           2h 42m (162 min) 3h 6m (186 min) (director's cut) 3h 35m (215 min) (Ultimate Cut)
            #           55m
            match = re.search(r'\((\d+)\s*(?:min|m)\)\s*(?:\((.+?)\)|)', li_text)
            if not match:
                match = re.search(r'^(\d+)\s*(?:min|m)$', li_text)

            if match:
                groups = match.groups()
                minutes = int(groups[0])
                if len(groups) >= 2:
                    key = utils.string.capitalize(groups[1]) if groups[1] else 'default'
                else:
                    key = 'default'

                if not any(regex.search(key) for regex in self._ignored_runtimes_keys):
                    runtimes[key] = minutes

        return runtimes

    async def summary(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/')

            # Get summary from the top
            candidates = (
                # Old website design
                soup.find(class_='summary_text'),
                # New website design
                soup.find(attrs={'data-testid': 'plot-xl'}),
            )
            for tag in candidates:
                if tag:
                    string = ''.join(tag.stripped_strings).strip()
                    link_texts = ('See full summary»', 'Read all', 'Add a Plot»')
                    if all(not string.endswith(text) for text in link_texts):
                        return string

            # Get summary from dedicated page
            soup = await self._get_soup(f'title/{id}/plotsummary')
            candidates = (
                # Old website design
                utils.html.get(soup.find(id='plot-summaries-content'), 'li', 'p'),
                # New website design
                soup.find(attrs={'data-testid': 'sub-section-summaries'}),
            )
            for tag in candidates:
                if tag:
                    text = utils.html.as_text(tag)
                    return re.sub(r'—.*$', '', text).strip()

        return ''

    async def _title_original(self, id):
        # /title/tt.../releaseinfo/#akas: Find "(original title)" row in table
        # Example: https://www.imdb.com/title/tt1405737/releaseinfo/#akas
        soup = await self._get_soup(f'title/{id}/releaseinfo')
        for tag in soup.select('.ipc-metadata-list__item > .ipc-metadata-list-item__label'):
            text = utils.html.as_text(tag)
            if 'original' in text:
                return utils.html.as_text(tag.next_sibling)

        return ''

    async def _titles_english(self, id):
        soup = await self._get_soup(f'title/{id}/releaseinfo')

        # Get the main (usually English) title from the <head>
        tag = soup.select_one('head meta[property="og:title"]')
        if tag:
            match = re.search(r'^(.*)\s+\(.*?\)\s*', tag['content'])
            if match:
                return [match.group(1).strip()]

        return ['']

    async def type(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/reference')

            tag = soup.find(class_='ipl-inline-list')
            text = ' '.join(reversed(tuple(tag.stripped_strings))).lower()

            if 'episode' in text:
                return ReleaseType.episode
            elif 'tv series' in text:
                return ReleaseType.season
            elif re.search(r'tv mini[- ]series', text):
                return ReleaseType.season
            elif 'tv movie' in text:
                return ReleaseType.movie
            elif 'movie' in text:
                return ReleaseType.movie
            elif 'short' in text:
                return ReleaseType.movie
            elif 'video' in text:
                return ReleaseType.movie

        return ReleaseType.unknown

    async def url(self, id):
        if id:
            return f'{self._url_base.rstrip("/")}/title/{id}'
        return ''

    async def year(self, id):
        if id:
            soup = await self._get_soup(f'title/{id}/reference')

            tag = soup.find(class_='titlereference-title-year')
            if tag:
                text = ' '.join(reversed(tuple(tag.stripped_strings))).lower()
                match = re.search(r'\b(\d{4})\b', text)
                if match:
                    return match.group(1)

        return ''


class _ImdbSearchResult(common.SearchResult):
    def __init__(self, *, imdb_api, soup=None, cast=None, countries=None,
                 directors=None, id=None, genres=None, summary=None, title=None,
                 title_english=None, title_original=None, type=None, url=None,
                 year=None):
        self._imdb_api = imdb_api
        soup = soup or utils.html.parse('')
        id = id or self._get_id(soup)
        return super().__init__(
            cast=cast or functools.partial(imdb_api.cast, id),
            countries=countries or functools.partial(imdb_api.countries, id),
            directors=directors or functools.partial(imdb_api.directors, id),
            genres=genres or functools.partial(imdb_api.genres, id),
            id=id or self._get_id(soup),
            summary=summary or self._get_summary(soup),
            title=title or self._get_title(soup),
            title_english=title_english or functools.partial(imdb_api.title_english, id),
            title_original=title_original or functools.partial(imdb_api.title_original, id),
            type=type or self._get_type(soup),
            url=url or self._get_url(soup),
            year=year or self._get_year(soup),
        )

    def _get_id(self, soup):
        a_tag = soup.find('a')
        if a_tag:
            href = a_tag.get('href')
            return re.sub(r'^.*/([t0-9]+)/.*$', r'\1', href)
        return ''

    def _get_summary(self, soup):
        summary = ''

        # Old website design
        tags = soup.find_all(class_='text-muted')
        if len(tags) >= 3:
            strings = tuple(tags[2].strings)
            if strings and 'Add a Plot' not in strings:
                summary = (''.join(strings) or '').strip()

        # Look for "See full summary" link. Preceding text is summary.
        if not summary:
            summary_link = soup.find('a', string=re.compile(r'(?i:full\s+summary)'))
            if summary_link:
                summary_tag = summary_link.parent
                if summary_tag:
                    summary = ''.join(summary_tag.strings)

        summary = re.sub(r'(?i:see full summary).*', '', summary).strip()
        summary = re.sub(r'\s*\.\.\.\s*$', '…', summary)

        # New website design
        if not summary:
            tag = soup.find('div', attrs={'class': 'ipc-html-content', 'role': 'presentation'})
            if tag:
                return tag.string.strip()

        return summary

    def _get_title(self, soup):
        # Old website design
        tag = soup.find(attrs={'class': 'lister-item-header'})
        if tag:
            a_tag = tag.find('a')
            if a_tag:
                return a_tag.string.strip()

        # New design
        tag = soup.find(attrs={'class': 'ipc-title__text'})
        if tag:
            title = tag.string.strip()
            # Search result number is prepended to title
            title = re.sub(r'^\d+\.', '', title)
            return title.strip()

        return ''

    def _get_type(self, soup):
        tag = soup.find(attrs={'class': 'dli-title-type-data'})
        if tag:
            text = tag.string.strip().lower()
            if any(name in text for name in ('movie', 'video', 'short')):
                return ReleaseType.movie
            elif 'series' in text:
                return ReleaseType.series

        # Sometimes there is no indication of type in search results. It seems
        # this only happens to movies.
        return ReleaseType.movie

    def _get_url(self, soup):
        id = self._get_id(soup)
        if id:
            return f'{ImdbApi._url_base}/title/{id}'
        return ''

    def _get_year(self, soup):
        # Old website design
        try:
            year = soup.find(class_='lister-item-year').string or ''
        except AttributeError:
            pass
        else:
            # Year may be preceded by roman number
            year = re.sub(r'\([IVXLCDM]+\)\s*', '', year)
            # Remove parentheses
            year = year.strip('()')
            # Possible formats:
            # - YYYY
            # - YYYY–YYYY  ("–" is NOT "U+002D / HYPHEN-MINUS" but "U+2013 / EN DASH")
            # - YYYY–
            try:
                return str(int(year[:4]))
            except (ValueError, TypeError):
                pass

        # New website design
        # Find the first four-digit number. This covers movies as well as series
        # in "YYYY-YYYY" and "YYYY-" format.
        year_regex = re.compile(r'(\d{4})')
        tags = soup.find_all(attrs={'class': 'dli-title-metadata-item'})
        for tag in tags:
            text = tag.string.strip()
            match = year_regex.search(text)
            if match:
                year = match.group(1)
                return year

        return ''



import re
from unittest.mock import Mock, call

import pytest

from upsies import errors
from upsies.utils import release
from upsies.utils.types import ReleaseType


@pytest.mark.parametrize(
    argnames='strict, exception, exp_exception',
    argvalues=(
        (False, None, None),
        (False, errors.SceneAbbreviatedFilenameError('foo'), None),
        (True, errors.SceneAbbreviatedFilenameError('foo'),
         errors.ContentError(errors.SceneAbbreviatedFilenameError('foo'))),
        (True, None, None),
    ),
    ids=lambda v: str(v),
)
def test_strict_argument(strict, exception, exp_exception, mocker):
    mocker.patch('upsies.utils.predbs.assert_not_abbreviated_filename', side_effect=exception)
    if exp_exception:
        exp_msg = re.escape(str(exp_exception))
        with pytest.raises(type(exp_exception), match=rf'^{exp_msg}$'):
            release.ReleaseInfo('foo.mkv', strict=strict)
    else:
        ri = release.ReleaseInfo('foo.mkv', strict=strict)
        assert ri['title'] == 'foo'


def test_getting_known_key(mocker):
    ri = release.ReleaseInfo('foo.mkv')
    mocker.patch.object(ri, '_get_title', Mock(return_value='Mocked Title'))
    for _ in range(3):
        assert ri['title'] == 'Mocked Title'
    assert ri._get_title.call_args_list == [call()]

def test_getting_unknown_key():
    ri = release.ReleaseInfo('foo.mkv')
    with pytest.raises(KeyError, match=r"^'foo'$"):
        ri['foo']


def test_setting_known_key():
    ri = release.ReleaseInfo('foo.mkv')
    assert ri['title'] == 'foo'
    ri['title'] = ''
    assert ri['title'] == ''

def test_setting_unknown_key():
    ri = release.ReleaseInfo('foo.mkv')
    with pytest.raises(KeyError, match=r"^'foo'$"):
        ri['foo'] = 'bar'
    assert 'foo' not in ri


def test_deleting_unknown_key():
    ri = release.ReleaseInfo('foo.mkv')
    with pytest.raises(KeyError, match=r"^'foo'$"):
        del ri['foo']

def test_deleting_known_key():
    ri = release.ReleaseInfo('foo.mkv')
    assert ri['title'] == 'foo'
    del ri['title']
    assert ri['title'] == 'foo'


def test_iter():
    ri = release.ReleaseInfo('foo.mkv')
    assert set(ri) == {
        'type',
        'title',
        'aka',
        'country',
        'year',
        'episodes',
        'episode_title',
        'date',
        'edition',
        'resolution',
        'service',
        'source',
        'audio_codec',
        'audio_channels',
        'hdr_format',
        'video_codec',
        'group',
        'has_commentary',
    }

def test_len():
    ri = release.ReleaseInfo('foo.mkv')
    assert len(ri) == 18


def test_repr():
    ri = release.ReleaseInfo('path/to/foo.mkv')
    assert repr(ri) == "ReleaseInfo('path/to/foo.mkv')"


def test_copy():
    ri = release.ReleaseInfo('path/to/The Foo 2000 DC 1080p NF WEB-DL AAC2.0 DoVi H.264-ASDF.mkv')
    cp = ri.copy()
    cp['video_codec'] = 'XviD'
    cp['audio_codec'] = 'MP3'
    cp['edition'].append('Uncut')
    assert dict(ri) == {
        'type': ReleaseType.movie,
        'title': 'The Foo',
        'aka': '',
        'country': '',
        'year': '2000',
        'episodes': {},
        'episode_title': '',
        'date': '',
        'edition': ["Director's Cut"],
        'has_commentary': False,
        'resolution': '1080p',
        'service': 'NF',
        'source': 'WEB-DL',
        'audio_codec': 'AAC',
        'audio_channels': '2.0',
        'hdr_format': 'DV',
        'video_codec': 'H.264',
        'group': 'ASDF',
    }
    assert dict(cp) == {
        'type': ReleaseType.movie,
        'title': 'The Foo',
        'aka': '',
        'country': '',
        'year': '2000',
        'episodes': {},
        'episode_title': '',
        'date': '',
        'edition': ["Director's Cut", 'Uncut'],
        'has_commentary': False,
        'resolution': '1080p',
        'service': 'NF',
        'source': 'WEB-DL',
        'audio_codec': 'MP3',
        'audio_channels': '2.0',
        'hdr_format': 'DV',
        'video_codec': 'XviD',
        'group': 'ASDF',
    }
    assert ri.path == 'path/to/The Foo 2000 DC 1080p NF WEB-DL AAC2.0 DoVi H.264-ASDF.mkv'
    assert cp.path == 'path/to/The Foo 2000 DC 1080p NF WEB-DL AAC2.0 DoVi H.264-ASDF.mkv'


def test_episodes_is_singleton():
    ri = release.ReleaseInfo('foo.mkv')
    episodes_id = id(ri['episodes'])
    ri['episodes'] = {1: (2, 3)}
    assert id(ri['episodes']) == episodes_id


@pytest.mark.parametrize(
    argnames=('release_name', 'exp_params'),
    argvalues=(
        ('The Foo S01 foo bar baz', 'foo bar baz'),
        ('The Foo S01E01 foo bar baz', 'foo bar baz'),
        ('The Foo S01E01E02 foo bar baz', 'foo bar baz'),
        ('The Foo S01E01E02E03 foo bar baz', 'foo bar baz'),
        ('The Foo E9S01E01S03E02 foo bar baz', 'foo bar baz'),
        ('The Foo Season 1 foo bar baz', 'foo bar baz'),
        ('The Foo Season1 foo bar baz', 'foo bar baz'),
        ('The Foo Season 1 Episode 1 foo bar baz', 'foo bar baz'),
        ('The Foo Season 1 Episode 1 Episode 2 foo bar baz', 'foo bar baz'),
        ('The Foo Season 1 Season 2 Episode 1 Episode 2 foo bar baz', 'foo bar baz'),
        ('The Foo Season1 Episode1 foo bar baz', 'foo bar baz'),
        ('The Foo Season1 Episode1 Episode2 foo bar baz', 'foo bar baz'),
        ('The Foo foo bar baz', 'The Foo foo bar baz'),
    ),
    ids=lambda v: str(v),
)
def test_release_name_params(release_name, exp_params):
    for name, exp in ((release_name, exp_params),
                      (release_name.replace(' ', '.'), exp_params.replace(' ', '.'))):
        for path in (f'path/to/{name}.mkv', f'path/to/{name}/asdf.mkv'):
            print(repr(path), repr(exp))
            info = release.ReleaseInfo(path)
            assert info.release_name_params == exp

def test_release_name_params_default():
    assert release.ReleaseInfo('path/to/Foo.mkv').release_name_params == 'Foo'
    assert release.ReleaseInfo('Foo.mkv').release_name_params == 'Foo'
    assert release.ReleaseInfo('Foo').release_name_params == 'Foo'


def assert_info(release_name,
                type=ReleaseType.unknown, title='', aka='', country='', year='', episodes={}, date='',
                edition=[], resolution='', service='', source='',
                audio_codec='', audio_channels='',
                hdr_format='', video_codec='', group='', has_commentary=False):
    # Test space- and dot-separated release name
    for rn in (release_name, release_name.replace(' ', '.')):
        if '/' not in rn:
            # Test release name in file name and in parent directory name
            paths = (f'foo/{rn}.mkv', f'{rn}/foo.mkv')
        else:
            # Release name is already coming as path and grandparent directory
            # is not used for parsing
            paths = (f'foo/{rn}.mkv',)

        for path in paths:
            info = release.ReleaseInfo(path)
            assert info['type'] == type
            assert info['title'] == title
            assert info['aka'] == aka
            assert info['country'] == country
            assert info['year'] == year
            assert info['date'] == date
            assert info['episodes'] == episodes
            assert info['edition'] == edition
            assert info['resolution'] == resolution
            assert info['service'] == service
            assert info['source'] == source
            assert info['audio_codec'] == audio_codec
            assert info['audio_channels'] == audio_channels
            assert info['hdr_format'] == hdr_format
            assert info['video_codec'] == video_codec
            assert info['group'] == group
            assert info['has_commentary'] == has_commentary


@pytest.mark.parametrize('release_name, expected', (
    ('The Foo 2000 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '2000',
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo UK 2000 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.movie, 'title': 'The Foo', 'country': 'UK', 'year': '2000',
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo S01E04S02 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'year': '', 'episodes': {'1': ['4'], '2': []},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo S01 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.season, 'title': 'The Foo', 'year': '', 'episodes': {'1': []},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo US S01 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.season, 'title': 'The Foo', 'country': 'US', 'year': '', 'episodes': {'1': []},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo 2000 S01 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.season, 'title': 'The Foo', 'year': '2000', 'episodes': {'1': []},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo 2000 S01E02 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'year': '2000', 'episodes': {'1': ['2']},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo 2000 S01E02S03E04 1080p NF WEB-DL AAC2.0 H.264-ASDF',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'year': '2000',
      'episodes': {'1': ['2'], '3': ['4']},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo S01 1080p NF WEB-DL AAC2.0 H.264-ASDF/asdf-foos01e02.mkv',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'episodes': {'1': ['2']},
      'resolution': '1080p', 'service': 'NF', 'source': 'WEB-DL',
      'audio_codec': 'AAC', 'audio_channels': '2.0', 'video_codec': 'H.264', 'group': 'ASDF'}),
    ('The Foo 1984-09-23 1080p BluRay DTS x264-ASDF',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'date': '1984-09-23',
      'resolution': '1080p', 'source': 'BluRay',
      'audio_codec': 'DTS', 'video_codec': 'x264', 'group': 'ASDF'}),
    ('The Foo 23.09.1984 1080p BluRay DTS x264-ASDF',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'date': '1984-09-23',
      'resolution': '1080p', 'source': 'BluRay',
      'audio_codec': 'DTS', 'video_codec': 'x264', 'group': 'ASDF'}),
))
def test_type_and_year_season_and_episode(release_name, expected):
    assert_info(release_name, **expected)


@pytest.mark.parametrize('release_name, expected', (
    ('The Foo 1984 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '1984',
      'resolution': '1080p', 'service': '', 'source': 'BluRay',
      'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
    ('1984 1984 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': '1984', 'year': '1984',
      'resolution': '1080p', 'service': '', 'source': 'BluRay',
      'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
    ('1984 2000 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': '1984', 'year': '2000',
      'resolution': '1080p', 'service': '', 'source': 'BluRay',
      'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
))
def test_title_and_year(release_name, expected):
    assert_info(release_name, **expected)


@pytest.mark.parametrize('release_name, expected', (
    ('The Foo 1984-05-25 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'date': '1984-05-25',
      'resolution': '1080p', 'service': '', 'source': 'BluRay',
      'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
))
def test_date(release_name, expected):
    assert_info(release_name, **expected)


@pytest.mark.parametrize('release_name, expected', (
    ('The Foo - The Bar 1984 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': 'The Foo - The Bar', 'year': '1984',
      'resolution': '1080p', 'service': '', 'source': 'BluRay',
      'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
    ('The Foo - The Bar - The Baz 1984 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': 'The Foo - The Bar - The Baz', 'year': '1984',
      'resolution': '1080p', 'service': '', 'source': 'BluRay',
      'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
    ('1984 AKA Nineteen Eighty-Four 1984 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': '1984', 'aka': 'Nineteen Eighty-Four', 'year': '1984',
      'resolution': '1080p', 'service': '',
      'source': 'BluRay', 'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
    ('1984 - Foo AKA Nineteen Eighty-Four 1984 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': '1984 - Foo', 'aka': 'Nineteen Eighty-Four', 'year': '1984',
      'resolution': '1080p', 'service': '',
      'source': 'BluRay', 'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
    ('Foo - Bar AKA Baz 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': 'Foo - Bar', 'aka': 'Baz', 'year': '',
      'resolution': '1080p', 'service': '',
      'source': 'BluRay', 'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}),
))
def test_title_and_alternative_title(release_name, expected):
    assert_info(release_name, **expected)


edition_samples = (
    ('The Foo 2000 Alternative Cut 1080p BluRay DTS x264-ASDF', {'edition': ['Alternative Cut']}),
    ('The Foo 2000 Alternative 1080p BluRay DTS x264-ASDF', {'edition': ['Alternative Cut']}),
    ("The Foo 2000 Collector's Edition 1080p BluRay DTS x264-ASDF", {'edition': ["Collector's Edition"]}),
    ('The Foo 2000 Collector 1080p BluRay DTS x264-ASDF', {'edition': ["Collector's Edition"]}),
    ('The Foo 2000 Criterion Collection 1080p BluRay DTS x264-ASDF', {'edition': ['Criterion Collection']}),
    ('The Foo 2000 Criterion 1080p BluRay DTS x264-ASDF', {'edition': ['Criterion Collection']}),
    ('The Foo 2000 CC 1080p BluRay DTS x264-ASDF', {'edition': ['Criterion Collection']}),
    ('The Foo 2000 Deluxe 1080p BluRay DTS x264-ASDF', {'edition': ['Deluxe Edition']}),
    ("The Foo 2000 Director's Cut 1080p BluRay DTS x264-ASDF", {'edition': ["Director's Cut"]}),
    ('The Foo 2000 DC 1080p BluRay DTS x264-ASDF', {'edition': ["Director's Cut"]}),
    # TODO: Enable when support was added to guessit:
    # ("The Foo 2000 Director's Definitive Cut 1080p BluRay DTS x264-ASDF", {'edition': ["Director's Definitive Cut"]}),
    ('The Foo 2000 DDC 1080p BluRay DTS x264-ASDF', {'edition': ["Director's Definitive Cut"]}),
    ('The Foo 2000 Extended Cut 1080p BluRay DTS x264-ASDF', {'edition': ['Extended Cut']}),
    ('The Foo 2000 Extended 1080p BluRay DTS x264-ASDF', {'edition': ['Extended Cut']}),
    ('The Foo 2000 IMAX 1080p BluRay DTS x264-ASDF', {'edition': ['IMAX']}),
    ('The Foo 2000 Limited 1080p BluRay DTS x264-ASDF', {'edition': ['Limited']}),
    ('The Foo 2000 Remastered 1080p BluRay DTS x264-ASDF', {'edition': ['Remastered']}),
    ('The Foo 2000 restored 1080p BluRay DTS x264-ASDF', {'edition': ['Restored']}),
    ('The Foo 2000 4k ReMastered 1080p BluRay DTS x264-ASDF', {'edition': ['4k Remastered']}),
    ('The Foo 2000 4K RESTORED 1080p BluRay DTS x264-ASDF', {'edition': ['4k Restored']}),
    ('The Foo 2000 Special Edition 1080p BluRay DTS x264-ASDF', {'edition': ['Special Edition']}),
    # TODO: Enable when support was added to guessit:
    # ('The Foo 2000 Special 1080p BluRay DTS x264-ASDF', {'edition': ['Special Edition']}),
    ('The Foo 2000 Theatrical Cut 1080p BluRay DTS x264-ASDF', {'edition': ['Theatrical Cut']}),
    ('The Foo 2000 Theatrical 1080p BluRay DTS x264-ASDF', {'edition': ['Theatrical Cut']}),
    ('The Foo 2000 Ultimate Edition 1080p BluRay DTS x264-ASDF', {'edition': ['Ultimate Cut']}),
    # TODO: Enable when support was added to guessit:
    # ('The Foo 2000 Ultimate Cut 1080p BluRay DTS x264-ASDF', {'edition': ['Ultimate Cut']}),
    # ('The Foo 2000 Ultimate 1080p BluRay DTS x264-ASDF', {'edition': ['Ultimate Cut']}),
    ('The Foo 2000 Uncensored 1080p BluRay DTS x264-ASDF', {'edition': ['Uncensored']}),
    ('The Foo 2000 Uncut 1080p BluRay DTS x264-ASDF', {'edition': ['Uncut']}),
    ('The Foo 2000 Unrated 1080p BluRay DTS x264-ASDF', {'edition': ['Unrated']}),
    ('The Foo 2000 1080p PROPER BluRay DTS x264-ASDF', {'edition': ['PROPER']}),
    ('The Foo 2000 Repack 1080p BluRay DTS x264-ASDF', {'edition': ['REPACK']}),
    ('The Foo 2000 1080p Repack2 BluRay DTS x264-ASDF', {'edition': ['REPACK2']}),
    ('The Foo 2000 1080p Dual Audio BluRay DTS x264-ASDF', {'edition': ['Dual Audio']}),
    ('The Foo 2000 DC Uncut 1080p BluRay DTS x264-ASDF', {'edition': ["Director's Cut", 'Uncut']}),
    ('The Foo 2000 1080p Hybrid Uncut BluRay DTS x264-ASDF', {'edition': ['Uncut'],
                                                              'source': 'Hybrid BluRay'}),
    ('The Foo 2000 1080p Hybrid Unrated DC BluRay DTS x264-ASDF', {'edition': ['Unrated', "Director's Cut"],
                                                                   'source': 'Hybrid BluRay'}),
    ('The Foo Extended 2000 1080p Hybrid Uncut BluRay DTS x264-ASDF', {'edition': ['Extended Cut', 'Uncut'],
                                                                       'source': 'Hybrid BluRay'}),
    ('The Foo 2000 OAR 1080p BluRay DTS x264-ASDF', {'edition': ['OAR']}),
    ('The Foo 2000 OM 1080p BluRay DTS x264-ASDF', {'edition': ['Open Matte']}),
    ('The Foo 2000 Open Matte 1080p BluRay DTS x264-ASDF', {'edition': ['Open Matte']}),
    ('The Foo 2000 1080p BluRay 2in1 DTS x264-ASDF', {'edition': ['2in1']}),
)
@pytest.mark.parametrize('release_name, exp_values', edition_samples)
def test_edition(release_name, exp_values):
    expected = {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '2000',
                'resolution': '1080p', 'source': 'BluRay',
                'audio_codec': 'DTS', 'video_codec': 'x264', 'group': 'ASDF'}
    expected.update(exp_values)
    assert_info(release_name, **expected)


source_samples = (
    ('DVDRip', 'DVDRip'), ('dvdrip', 'DVDRip'), ('dvd rip', 'DVDRip'), ('DVDRip hybrid', 'Hybrid DVDRip'),
    ('TVRip', 'TVRip'), ('tvrip', 'TVRip'), ('tv-rip', 'TVRip'),
    ('BluRay', 'BluRay'), ('bluray', 'BluRay'), ('Blu-ray', 'BluRay'),
    ('UHD BluRay', 'UHD BluRay'), ('uhd bluray', 'UHD BluRay'), ('Ultra HD Blu-ray', 'UHD BluRay'),
    ('Blu-ray Hybrid REMUX', 'Hybrid BluRay Remux'),
    ('WEB-DL', 'WEB-DL'), ('WEBDL', 'WEB-DL'), ('web-dl', 'WEB-DL'), ('webdl', 'WEB-DL'),
    ('WEBRip', 'WEBRip'), ('WEBRIP', 'WEBRip'), ('web-rip', 'WEBRip'), ('webrip', 'WEBRip'),
    ('WEB-DL Remux', 'WEB-DL'), ('Remux WEB-DL', 'WEB-DL'), ('WEBRip Remux', 'WEBRip'), ('Remux WEBRip', 'WEBRip'),
    ('WEB', 'WEB-DL'), ('Web', 'WEB-DL'), ('web', 'WEB-DL'),
    ('DVD', 'DVD'), ('DVD9', 'DVD9'), ('DVD5', 'DVD5'),
    ('DVD Remux', 'DVD Remux'), ('Remux DVD', 'DVD Remux'),
    # An episode title can contain another source and guessit doesn't have a
    # problem with both "WEB" and "Blu-ray" as a source.
    ('WEB-DL tc', 'WEB-DL'), ('BluRay DVD', 'BluRay'),
    # TODO: Blu-ray images. guessit doesn't support it. Can we just look for
    #       "AVC" and "video_codec=H.264"?
)
@pytest.mark.parametrize('source, exp_source', source_samples)
def test_source(source, exp_source):
    release_name = f'The Foo 1984 1080p {source} DTS-ASDF'
    expected = {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '1984',
                'resolution': '1080p', 'service': '', 'source': exp_source,
                'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}
    assert_info(release_name, **expected)


extended_source_samples = (
    ('BluRay REMUX', 'BluRay Remux'),
    ('remux BluRay', 'BluRay Remux'),
    ('DVD Remux', 'DVD Remux'),
    ('DVDRip Remux', 'DVDRip'),
    ('Hybrid BluRay', 'Hybrid BluRay'),
    ('BluRay hybrid', 'Hybrid BluRay'),
    ('HYBRID BluRay REMUX', 'Hybrid BluRay Remux'),
    ('Bluray hybrid remux', 'Hybrid BluRay Remux'),
    ('Blu-Ray REMUX hybrid', 'Hybrid BluRay Remux'),
    ('remux BluRay Hybrid', 'Hybrid BluRay Remux'),
)
@pytest.mark.parametrize('source, exp_source', extended_source_samples)
def test_extended_source(source, exp_source):
    release_name = f'The Foo 1984 1080p {source} DTS-ASDF'
    expected = {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '1984',
                'resolution': '1080p', 'service': '', 'source': exp_source,
                'audio_codec': 'DTS', 'audio_channels': '', 'group': 'ASDF'}
    assert_info(release_name, **expected)


service_samples = (
    ('NF', 'NF', []), ('Netflix', 'NF', []),
    ('AMZN', 'AMZN', []), ('Amazon', 'AMZN', []),
    ('APTV', 'ATVP', []),
    ('BBC', 'BBC', []),
    ('HBO', 'HBO', []),
    ('HMAX', 'HMAX', []),
    ('HULU', 'HULU', []), ('Hulu', 'HULU', []),
    ('iT', 'iT', []), ('iTunes', 'iT', []),
    ('VUDU', 'VUDU', []),
    ('CRKL', 'CRKL', []),
    ('ALL4', 'ALL4', []),
    ('4oD', 'ALL4', []),
    # Not streaming services
    ('OAR', '', ['OAR']),
)
@pytest.mark.parametrize('service, service_abbrev, edition', service_samples)
def test_service_is_abbreviated(service, service_abbrev, edition):
    release_name = f'The Foo S10 1080p {service} WEB-DL AAC H.264-ASDF'
    expected = {'type': ReleaseType.season, 'title': 'The Foo', 'year': '', 'episodes': {'10': []},
                'edition': edition, 'resolution': '1080p', 'service': service_abbrev, 'source': 'WEB-DL',
                'audio_codec': 'AAC', 'audio_channels': '', 'video_codec': 'H.264', 'group': 'ASDF'}
    assert_info(release_name, **expected)


audio_codec_samples = (
    ('DD', 'DD'), ('AC-3', 'DD'), ('AC3', 'DD'),
    ('DDP', 'DDP'), ('DD+', 'DDP'), ('E-AC-3', 'DDP'), ('E-AC3', 'DDP'), ('EAC-3', 'DDP'),
    ('Atmos', 'Atmos'), ('DD+ Atmos', 'Atmos'), ('DDP Atmos', 'Atmos'),
    ('TrueHD', 'TrueHD'), ('True-HD', 'TrueHD'), ('Dolby TrueHD', 'TrueHD'),
    ('DD TrueHD', 'TrueHD'), ('DDP True-HD', 'TrueHD'), ('Dolby TrueHD', 'TrueHD'),
    ('TrueHD Atmos', 'TrueHD Atmos'), ('Dolby True-HD Atmos', 'TrueHD Atmos'),
    ('DTS', 'DTS'),
    ('AAC', 'AAC'),
    ('FLAC', 'FLAC'),
    ('MP3', 'MP3'),
    ('Opus', 'Opus'),
    (None, ''),
)
@pytest.mark.parametrize('audio_codec, audio_codec_abbrev', audio_codec_samples)
def test_audio_codec(audio_codec, audio_codec_abbrev):
    release_name = f'The Foo S04 1080p WEB-DL {audio_codec} H.264-ASDF'
    expected = {'type': ReleaseType.season, 'title': 'The Foo', 'year': '', 'episodes': {'4': []},
                'resolution': '1080p', 'source': 'WEB-DL',
                'audio_codec': audio_codec_abbrev, 'audio_channels': '',
                'video_codec': 'H.264', 'group': 'ASDF'}
    assert_info(release_name, **expected)


audio_channels_samples = (
    ('MP3', 'MP3', ''),
    ('FLAC 1.0', 'FLAC', '1.0'),
    ('AAC2.0', 'AAC', '2.0'),
    ('AAC 2.0', 'AAC', '2.0'),
    ('AC3 5.1', 'DD', '5.1'),
    ('DD+ 5.1', 'DDP', '5.1'),
    ('Dolby Digital 7.1', 'DD', '7.1'),
    ('Opus 7.1', 'Opus', '7.1'),
)
@pytest.mark.parametrize('audio, audio_codec, audio_channels', audio_channels_samples)
def test_audio_channels(audio, audio_codec, audio_channels):
    release_name = f'The Foo 2000 1080p WEB-DL {audio} H.264-ASDF'
    expected = {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '2000',
                'resolution': '1080p', 'source': 'WEB-DL',
                'audio_codec': audio_codec, 'audio_channels': audio_channels,
                'video_codec': 'H.264', 'group': 'ASDF'}
    assert_info(release_name, **expected)


hdr_format_samples = (
    # TODO: guessit fails to detect audio_codec if {hdr} is between audio_codec and video_codec.
    # ('The Foo 2000 1080p DTS {hdr} x264-ASDF', {'year': '2000', 'resolution': '1080p',
    #                                             'audio_codec': 'DTS', 'video_codec': 'x264',
    #                                             'group': 'ASDF', 'type': ReleaseType.movie}),
    # TODO: If {hdr} is in fron of video_codec with no group, guessit thinks {hdr} is the group.
    # ('The Foo 2000 1080p {hdr} x264', {'year': '2000', 'resolution': '1080p',
    #                                    'video_codec': 'x264',
    #                                    'type': ReleaseType.movie}),
    ('The Foo 2000 1080p {hdr} x264-ASDF', {'year': '2000', 'resolution': '1080p',
                                            'video_codec': 'x264',
                                            'group': 'ASDF', 'type': ReleaseType.movie}),
    ('The Foo 2000 1080p Uncut {hdr} x264-ASDF', {'year': '2000', 'resolution': '1080p',
                                                  'edition': ['Uncut'],
                                                  'video_codec': 'x264',
                                                  'group': 'ASDF', 'type': ReleaseType.movie}),
)
@pytest.mark.parametrize('hdr_format, exp_hdr_format', (
    ('DV', 'DV'),
    ('DoVi', 'DV'),
    ('Dolby Vision', 'DV'),
    ('HDR10+', 'HDR10+'),
    ('HDR10', 'HDR10'),
    ('HDR', 'HDR'),
))
@pytest.mark.parametrize('release_name, exp_values', hdr_format_samples, ids=lambda v: str(v))
def test_hdr_format(release_name, exp_values, hdr_format, exp_hdr_format):
    exp_values['hdr_format'] = exp_hdr_format
    expected = {**{'title': 'The Foo'}, **exp_values}
    rn = release_name.format(hdr=hdr_format)
    assert_info(rn, **expected)


video_codec_samples = (
    ('The Foo 2000 1080p DTS {vc}-ASDF', {'year': '2000', 'resolution': '1080p', 'audio_codec': 'DTS',
                                          'group': 'ASDF', 'type': ReleaseType.movie}),
    ('The Foo 2000 1080p DTS {vc}', {'year': '2000', 'resolution': '1080p', 'audio_codec': 'DTS',
                                     'type': ReleaseType.movie}),
    ('The Foo S03 1080p DTS {vc}-ASDF', {'episodes': {'3': []}, 'resolution': '1080p', 'audio_codec': 'DTS',
                                         'group': 'ASDF', 'type': ReleaseType.season}),
    ('The Foo S03 1080p DTS {vc}', {'episodes': {'3': []}, 'resolution': '1080p', 'audio_codec': 'DTS',
                                    'type': ReleaseType.season}),
    ('The Foo S03 {vc}-ASDF', {'episodes': {'3': []}, 'group': 'ASDF', 'type': ReleaseType.season}),
    ('The Foo S03 {vc}', {'episodes': {'3': []}, 'type': ReleaseType.season}),
    ('The Foo {vc}-ASDF', {'group': 'ASDF', 'type': ReleaseType.movie}),
    ('The Foo {vc}', {'type': ReleaseType.movie}),
)
@pytest.mark.parametrize('video_codec, exp_video_codec', (
    ('x264', 'x264'),
    ('X264', 'x264'),
    ('H.264', 'H.264'),
    ('H264', 'H.264'),
    ('x265', 'x265'),
    ('X265', 'x265'),
    ('H.265', 'H.265'),
    ('H265', 'H.265'),
))
@pytest.mark.parametrize('release_name, exp_values', video_codec_samples, ids=lambda v: str(v))
def test_video_codec(release_name, exp_values, video_codec, exp_video_codec):
    exp_values['video_codec'] = exp_video_codec
    expected = {**{'title': 'The Foo'}, **exp_values}
    rn = release_name.format(vc=video_codec)
    assert_info(rn, **expected)


group_samples = (
    ('The Foo 2000 1080p DTS x264-ASDF', 'ASDF'),
    ('The Foo 2000 1080p DTS x264-AsdF', 'AsdF'),
    ('The Foo 2000 1080p DTS x264-A-F', 'A-F'),
)
@pytest.mark.parametrize('release_name, group', group_samples)
def test_group(release_name, group):
    expected = {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '2000',
                'resolution': '1080p',
                'audio_codec': 'DTS', 'video_codec': 'x264', 'group': group}
    assert_info(release_name, **expected)


@pytest.mark.parametrize(
    argnames='release_name, exp_value',
    argvalues=(
        ('The Foo 2000 1080p Plus Comm DTS x264-ASDF', True),
        ('The Foo 2000 1080p Comm DTS x264-ASDF', False),
        ('The Foo 2000 1080p Commentary DTS x264-ASDF', True),
        ('The Foo 2000 1080p DTS x264-ASDF', False),
    ),
)
def test_has_commentary(release_name, exp_value):
    expected = {'type': ReleaseType.movie, 'title': 'The Foo', 'year': '2000',
                'resolution': '1080p', 'audio_codec': 'DTS', 'video_codec': 'x264',
                'group': 'ASDF', 'has_commentary': exp_value}
    assert_info(release_name, **expected)
    ri = release.ReleaseInfo(release_name)
    ri['has_commentary'] = ''
    assert ri['has_commentary'] is False
    ri['has_commentary'] = 1
    assert ri['has_commentary'] is True
    ri['has_commentary'] = None
    assert ri['has_commentary'] is exp_value


def test_path():
    assert release.ReleaseInfo('path/to/foo').path == 'path/to/foo'


@pytest.mark.parametrize('path, exp_release_name', (
    ('Foo.2017.720p.BluRay.x264-ASDF/asdf-foo.2017.720p.bluray.x264.mkv',
     'Foo 2017 720p BluRay x264-ASDF'),
    ('Bar.Bar.2013.720p.BluRay.x264-ASDF/asdf-barbar.mkv',
     'Bar Bar 2013 720p BluRay x264-ASDF'),
    ('path/to/Baz.2009.720p.BluRay.x264-ASDF/asdf-720p-baz.mkv',
     'Baz 2009 720p BluRay x264-asdf'),
    ('QuuX.2005.1080p.BluRay.x264-ASD/asd-q_u_u_x_x264_bluray.mkv',
     'QuuX 2005 1080p BluRay x264-ASD'),
    ('foo/foo.bar.1994.1080p.blu-ray.x264-baz.mkv',
     'foo bar 1994 1080p BluRay x264-baz'),
    ('foo/asdf-barbar.mkv',
     'asdf-barbar    -'),
))
def test_abbreviated_scene_filename(path, exp_release_name):
    ri = release.ReleaseInfo(path)
    release_name = '{title} {year} {resolution} {source} {video_codec}-{group}'.format(**ri)
    assert release_name == exp_release_name


@pytest.mark.parametrize('release_name, expected', (
    ('The Collector 2009 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': 'The Collector', 'year': '2009',
      'resolution': '1080p', 'source': 'BluRay', 'audio_codec': 'DTS', 'group': 'ASDF',
      'edition': []}),
    ("The Collector 2009 Collector's Edition 1080p BluRay DTS-ASDF",
     {'type': ReleaseType.movie, 'title': 'The Collector', 'year': '2009',
      'resolution': '1080p', 'source': 'BluRay', 'audio_codec': 'DTS', 'group': 'ASDF',
      'edition': ["Collector's Edition"]}),
    ('xXx 2002 1080p BluRay DTS-ASDF',
     {'type': ReleaseType.movie, 'title': 'xXx', 'year': '2002',
      'resolution': '1080p', 'source': 'BluRay', 'audio_codec': 'DTS', 'group': 'ASDF'}),
    ('path/to/The Foo - Season 01 - DVDRip.x264',
     {'type': ReleaseType.season, 'title': 'The Foo', 'episodes': {'1': []},
      'source': 'DVDRip', 'video_codec': 'x264'}),
    ('The Foo - Season 02 - DVDRip.x264/Episode 3 - Something.mkv',
     {'type': ReleaseType.episode, 'title': 'The Foo', 'episodes': {'2': ['3']},
      'source': 'DVDRip', 'video_codec': 'x264'}),
    ('Real.Life.2015.1080p.AMZN.WEB-DL.DD+5.1.H.264-ASDF',
     {'type': ReleaseType.movie, 'title': 'Real', 'year': '2015', 'resolution': '1080p',
      'service': 'AMZN', 'source': 'WEB-DL', 'audio_codec': 'DDP', 'audio_channels': '5.1',
      'video_codec': 'H.264', 'group': 'ASDF'}),
))
def test_special_case(release_name, expected):
    assert_info(release_name, **expected)



import re
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.trackers import nbl
from upsies.utils.http import Response


def test_name_attribute():
    assert nbl.NblTracker.name == 'nbl'


def test_label_attribute():
    assert nbl.NblTracker.label == 'NBL'


@pytest.mark.asyncio
async def test_login(mocker):
    tracker = nbl.NblTracker()
    await tracker._login()


@pytest.mark.asyncio
async def test_logout(mocker):
    tracker = nbl.NblTracker()
    await tracker._logout()


@pytest.mark.asyncio
async def test_get_announce_url(mocker):
    tracker = nbl.NblTracker(options={'announce_url': 'https://nbl.local:123/d34db33f/announce'})
    announce_url = await tracker.get_announce_url()
    assert announce_url == 'https://nbl.local:123/d34db33f/announce'


def test_get_upload_url(mocker):
    tracker = nbl.NblTracker(options={'upload_url': 'https://nbl.local/upload'})
    upload_url = tracker.get_upload_url()
    assert upload_url == 'https://nbl.local/upload'


@pytest.mark.asyncio
async def test_upload_without_api_key(mocker):
    tracker = nbl.NblTracker(
        options={
            'apikey': '',
            'upload_url': 'http://nbl.local/upload',
            'announce': 'http://nbl.local/announce',
        },
    )
    http_mock = mocker.patch('upsies.utils.http', Mock(
        get=AsyncMock(),
        post=AsyncMock(),
    ))
    get_upload_url_mock = mocker.patch.object(tracker, 'get_upload_url')
    tracker_jobs_mock = Mock()

    with pytest.raises(errors.RequestError, match=r'^No API key configured$'):
        await tracker.upload(tracker_jobs_mock)

    assert get_upload_url_mock.call_args_list == []
    assert http_mock.get.call_args_list == []
    assert http_mock.post.call_args_list == []


@pytest.mark.asyncio
async def test_upload_succeeds(mocker):
    tracker = nbl.NblTracker(options={'apikey': 'thisismyapikey'})
    http_mock = mocker.patch('upsies.trackers.nbl.tracker.http', Mock(
        get=AsyncMock(),
        post=AsyncMock(return_value=Response(
            text='''
            {
                "status": "success"
            }
            ''',
            bytes=b'irrelevant',
        )),
    ))
    mocker.patch.object(tracker, 'get_upload_url', return_value='http://upload.url')
    tracker_jobs_mock = Mock(
        post_data={
            'one': 'Hello',
            'two': '',
            'three': 123,
            'four': None,
            'five': 'World',
        },
        torrent_filepath='path/to/file.torrent',
    )

    return_value = await tracker.upload(tracker_jobs_mock)
    assert return_value == tracker_jobs_mock.torrent_filepath

    assert http_mock.get.call_args_list == []
    assert http_mock.post.call_args_list == [call(
        url='http://upload.url',
        cache=False,
        user_agent=True,
        data={
            'one': 'Hello',
            'three': '123',
            'five': 'World',
        },
        files={
            'file_input': {
                'file': 'path/to/file.torrent',
                'mimetype': 'application/x-bittorrent',
            },
        },
    )]


@pytest.mark.parametrize(
    argnames='response, exp_exception',
    argvalues=(
        (Response(text='{"status": "success"}', bytes=b'irrelevant'),
         None),
        (Response(text='{"status": "success?"}', bytes=b'irrelevant'),
         RuntimeError("Unexpected response: {'status': 'success?'}")),

        # NOTE: We expect error messages as JSON in error responses
        #       (RequestError) and success responses (Response).

        (errors.RequestError(msg='irrelevant', text='{"message": "Maybe!"}'),
         errors.RequestError('Upload failed: Maybe!')),
        (Response(text='{"message": "Maybe!"}', bytes=b'irrelevant'),
         errors.RequestError('Upload failed: Maybe!')),

        (errors.RequestError(msg='irrelevant', text='{"error": "No!"}'),
         errors.RequestError('Upload failed: No!')),
        (Response(text='{"error": "No!"}', bytes=b'irrelevant'),
         errors.RequestError('Upload failed: No!')),

        (errors.RequestError(msg='irrelevant', text='{"invalid": json]'),
         errors.RequestError('Upload failed: {"invalid": json]')),
        (Response(text='{"invalid": json]', bytes=b'irrelevant'),
         errors.RequestError("""Malformed JSON: '{"invalid": json]': Expecting value: line 1 column 13 (char 12)""")),

        (errors.RequestError(msg='irrelevant', text='{"foo": "Bar!"}'),
         RuntimeError("Unexpected response: {'foo': 'Bar!'}")),
        (Response(text='{"foo": "Bar!"}', bytes=b'irrelevant'),
         RuntimeError("Unexpected response: {'foo': 'Bar!'}")),
    ),
)
@pytest.mark.asyncio
async def test_upload_handles_errors(response, exp_exception, mocker):
    tracker = nbl.NblTracker(options={'apikey': 'thisismyapikey'})
    upload_url = 'http://mock.url/upload.php'
    http_mock = mocker.patch('upsies.trackers.nbl.tracker.http', Mock(
        get=AsyncMock(),
        post=AsyncMock(),
    ))
    if isinstance(response, Exception):
        http_mock.post.side_effect = response
    else:
        http_mock.post.return_value = response
    mocker.patch.object(tracker, 'get_upload_url', return_value=upload_url)
    tracker_jobs_mock = Mock(
        post_data={'one': 'Hello', 'two': 'World'},
        torrent_filepath='path/to/file.torrent',
    )

    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            await tracker.upload(tracker_jobs_mock)
    else:
        return_value = await tracker.upload(tracker_jobs_mock)
        assert return_value is tracker_jobs_mock.torrent_filepath

    assert http_mock.get.call_args_list == []
    assert http_mock.post.call_args_list == [call(
        url=upload_url,
        cache=False,
        user_agent=True,
        data={
            'one': 'Hello',
            'two': 'World',
        },
        files={
            'file_input': {
                'file': 'path/to/file.torrent',
                'mimetype': 'application/x-bittorrent',
            },
        },
    )]



import json
import re
from unittest.mock import AsyncMock, call

import pytest

from upsies import __project_name__, constants, errors
from upsies.utils import fs
from upsies.utils.http import Response
from upsies.utils.imghosts import ptpimg


def test_name():
    assert ptpimg.PtpimgImageHost.name == 'ptpimg'


def test_default_config():
    assert ptpimg.PtpimgImageHost.default_config == {
        **ptpimg.PtpimgImageHost.default_config_common,
        **{
            'apikey': '',
            'base_url': 'https://ptpimg.me',
        },
    }


def test_description():
    assert ptpimg.PtpimgImageHost.description == (
        'You need an API key to upload images.\n'
        '\n'
        '  1. Create an account: https://ptpimg.me/register.php\n'
        f'  2. Store your API key in {fs.tildify_path(constants.IMGHOSTS_FILEPATH)}:\n'
        '\n'
        f'       $ {__project_name__} set --fetch-ptpimg-apikey EMAIL PASSWORD'
    )


@pytest.mark.asyncio
async def test_upload_image_succeeds(mocker, tmp_path):
    post_mock = mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text='[{"code": "this_is_the_code", "ext": "png"}]',
        bytes=b'irrelevant',
    )))
    imghost = ptpimg.PtpimgImageHost(options={'apikey': 'f00'}, cache_directory=tmp_path)
    url = await imghost._upload_image('some/path.jpg')
    assert url == imghost.options['base_url'] + '/this_is_the_code.png'
    assert post_mock.call_args_list == [call(
        url=imghost.options['base_url'] + '/upload.php',
        cache=False,
        headers={
            'referer': imghost.options['base_url'] + '/index.php',
        },
        data={
            'api_key': 'f00',
        },
        files={
            'file-upload[0]': 'some/path.jpg',
        },
    )]

@pytest.mark.parametrize(
    argnames='json_response',
    argvalues=(
        'foo',
        123,
        [],
        [[]],
        [{}],
        [{'notcode': 'this_is_the_code', 'ext': 'png'}],
        [{'code': 'this_is_the_code', 'notext': 'png'}],
        [{'code': '', 'ext': 'png'}],
        [{'code': 'this_is_the_code', 'ext': ''}],
    ),
)
@pytest.mark.asyncio
async def test_upload_image_gets_unexpected_json(json_response, mocker, tmp_path):
    mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text=json.dumps(json_response),
        bytes=b'irrelevant',
    )))
    imghost = ptpimg.PtpimgImageHost(options={'apikey': 'f00'}, cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=rf'^Unexpected response: {re.escape(str(json_response))}$'):
        await imghost._upload_image('some/path.jpg')


@pytest.mark.asyncio
async def test_get_apikey_finds_apikey(mocker, tmp_path):
    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock())
    post_mock = mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text=(
            '<html>'
            '<body>'
            '<form>'
            '<input id="api_key" name="api_key" value="d00d1e" />'
            '</form>'
            '</body>'
            '</html>'
        ),
        bytes=b'irrelevant',
    )))
    imghost = ptpimg.PtpimgImageHost(options={}, cache_directory=tmp_path)
    apikey = await imghost.get_apikey('foo@localhost', 'hunter2')
    assert apikey == 'd00d1e'
    assert post_mock.call_args_list == [call(
        url=f'{imghost.options["base_url"]}/login.php',
        cache=False,
        data={
            'email': 'foo@localhost',
            'pass': 'hunter2',
            'login': '',
        },
    )]
    assert get_mock.call_args_list == [call(f'{imghost.options["base_url"]}/logout.php')]

@pytest.mark.asyncio
async def test_get_apikey_when_request_fails(mocker, tmp_path):
    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock())
    post_mock = mocker.patch('upsies.utils.http.post', AsyncMock(
        side_effect=errors.RequestError('no interwebs'),
    ))
    imghost = ptpimg.PtpimgImageHost(options={}, cache_directory=tmp_path)
    with pytest.raises(errors.RequestError, match=r'^no interwebs$'):
        await imghost.get_apikey('foo@localhost', 'hunter2')
    assert post_mock.call_args_list == [call(
        url=f'{imghost.options["base_url"]}/login.php',
        cache=False,
        data={
            'email': 'foo@localhost',
            'pass': 'hunter2',
            'login': '',
        },
    )]
    assert get_mock.call_args_list == []

@pytest.mark.asyncio
async def test_get_apikey_with_wrong_login(mocker, tmp_path):
    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock())
    post_mock = mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text=(
            '<html>'
            ' <body>'
            '  <div class="container">'
            '   <div class="page-header">'
            '    <div class="panel panel-danger">'
            '     <div class="panel-heading">'
            '      <i class="glyphicon glyphicon-remove">'
            '      </i>'
            '      <span>'
            '       Error!'
            '      </span>'
            '     </div>'
            '     <div class="panel-body">'
            '      <ul>'
            '       <li>'
            '        Credentials are incorrect. Please try again.'
            '       </li>'
            '      </ul>'
            '     </div>'
            '    </div>'
            '   </div>'
            '  </div>'
            ' </body>'
            '</html>'
        ),
        bytes=b'irrelevant',
    )))
    imghost = ptpimg.PtpimgImageHost(options={}, cache_directory=tmp_path)
    with pytest.raises(errors.RequestError, match=r'^Credentials are incorrect\. Please try again\.$'):
        await imghost.get_apikey('foo@localhost', 'hunter2')
    assert post_mock.call_args_list == [call(
        url=f'{imghost.options["base_url"]}/login.php',
        cache=False,
        data={
            'email': 'foo@localhost',
            'pass': 'hunter2',
            'login': '',
        },
    )]
    assert get_mock.call_args_list == [call(f'{imghost.options["base_url"]}/logout.php')]

@pytest.mark.asyncio
async def test_get_apikey_fails_to_find_apikey(mocker, tmp_path):
    get_mock = mocker.patch('upsies.utils.http.get', AsyncMock())
    post_mock = mocker.patch('upsies.utils.http.post', AsyncMock(return_value=Response(
        text='',
        bytes=b'irrelevant',
    )))
    imghost = ptpimg.PtpimgImageHost(options={}, cache_directory=tmp_path)
    with pytest.raises(RuntimeError, match=r'^Failed to find API key$'):
        await imghost.get_apikey('foo@localhost', 'hunter2')
    assert post_mock.call_args_list == [call(
        url=f'{imghost.options["base_url"]}/login.php',
        cache=False,
        data={
            'email': 'foo@localhost',
            'pass': 'hunter2',
            'login': '',
        },
    )]
    assert get_mock.call_args_list == [call(f'{imghost.options["base_url"]}/logout.php')]
